"""concept_id is string

Revision ID: 88214d86a9d1
Revises: 79ff53a30228
Create Date: 2023-01-18 13:17:21.700138

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '88214d86a9d1'
down_revision = '79ff53a30228'


def upgrade():
    with op.batch_alter_table('concept') as batch_op:
        batch_op.alter_column('concept_id', type_=sa.String, existing_type=sa.Integer)


def downgrade():
    # Drop concepts which have non-integer concept_id
    op.execute("DELETE FROM concept WHERE cast(cast(concept_id AS INTEGER) AS TEXT) != concept_id")
    with op.batch_alter_table('concept') as batch_op:
        batch_op.alter_column('concept_id', existing_type=sa.String, type_=sa.Integer)



import argparse
import datetime
import logging
import os
import xml.etree.ElementTree as ElementTree
from builtins import input
from os import listdir
from os.path import isfile

from pyramid.paster import bootstrap
from pyramid.paster import get_appsettings
from pyramid.paster import setup_logging

from atramhasis.errors import SkosRegistryNotFoundException

log = logging.getLogger(__name__)


def write_element_to_xml(filename, sitemap_dir, element):
    tree = ElementTree.ElementTree(element)
    file_path_name = os.path.join(sitemap_dir, filename)
    tree.write(file_path_name, encoding='utf-8', xml_declaration=True)


def create_sitemaps(settings, limit_per_deel, directory, env):
    base_url = settings.get("atramhasis.url")
    schemes_url = f"{base_url}/conceptschemes/{{}}"
    concepts_url = f"{schemes_url}/c/{{}}"

    request = env['request']

    if hasattr(request, 'skos_registry') and request.skos_registry is not None:
        skos_registry = request.skos_registry
    else:
        raise SkosRegistryNotFoundException()  # pragma: no cover

    scheme_urls = list()
    concept_urls = list()
    for p in skos_registry.get_providers():
        if any([not_shown in p.get_metadata()['subject']
                for not_shown in ['external', 'hidden']]):
            continue
        scheme_id = p.get_metadata()['id']
        scheme_urls.append(schemes_url.format(scheme_id))
        concept_urls.extend([concepts_url.format(scheme_id, x['id'])
                             for x in p.get_all()])

    create_deel_sitemaps(scheme_urls, limit_per_deel, directory, 'conceptschemes')
    create_deel_sitemaps(concept_urls, limit_per_deel, directory, 'concepts')

    create_index_sitemap(base_url, directory)


def create_deel_sitemaps(objecturls, limit_per_deel, sitemap_dir, name):
    """
    Sitemaps have a maximum amount of items. This method creates partial
    sitemaps with up to `limit_per_deel` items per file.
    """
    log.info("Beginning creation of sitemaps...")
    urlset = ElementTree.Element(
        "urlset", xmlns="http://www.sitemaps.org/schemas/sitemap/0.9"
    )
    sitemap_counter = 1

    for counter, objecturl in enumerate(objecturls, 1):
        url = ElementTree.SubElement(urlset, "url")
        ElementTree.SubElement(url, "loc").text = objecturl

        if counter % limit_per_deel == 0:
            filename = f'{name}_sitemap_deel_{sitemap_counter}.xml'
            log.info("Processed %s conceptschemes, writing %s", counter, filename)
            write_element_to_xml(filename, sitemap_dir, urlset)
            sitemap_counter += 1
            urlset = ElementTree.Element(
                "urlset", xmlns="http://www.sitemaps.org/schemas/sitemap/0.9"
            )
    if len(urlset):
        filename = f'{name}_sitemap_deel_{sitemap_counter}.xml'
        write_element_to_xml(filename, sitemap_dir, urlset)
    log.info(f"All {name} sitemaps created.")


def create_index_sitemap(base_url, directory):
    """Loop over all the created sitemaps, and create an index file."""
    log.info("Beginning creation of the final sitemap index...")
    list_sitemaps = [
        f for f in listdir(directory)
        if isfile(os.path.join(directory, f))
        and "sitemap" in f and "sitemap_index.xml" not in f
    ]
    sitemapindex = ElementTree.Element(
        "sitemapindex", xmlns="http://www.sitemaps.org/schemas/sitemap/0.9"
    )
    today = datetime.date.today().isoformat()

    for file_name in list_sitemaps:
        sitemap_static_url = f"{base_url}/sitemaps/{file_name}"
        sitemap_area = ElementTree.SubElement(sitemapindex, "sitemap")
        ElementTree.SubElement(sitemap_area, "loc").text = sitemap_static_url
        ElementTree.SubElement(sitemap_area, "lastmod").text = today

    write_element_to_xml("sitemap_index.xml", directory, sitemapindex)
    log.info("Sitemap index created.")


def main():
    parser = argparse.ArgumentParser(
        description="Process some command line arguments. ",
        usage="sitemap_generator development.ini#atramhasis "
              "--limit=1000")
    parser.add_argument('settings_file',
                        help="<The location of the settings file>")
    parser.add_argument("--limit", type=int,
                        help="range of objects in sitemap", default=50000)
    parser.add_argument("--no-input", action='store_true',
                        help="Don't stop script for user input")
    args, _ = parser.parse_known_args()

    config_uri = args.settings_file
    limit = args.limit
    setup_logging(config_uri)
    settings = get_appsettings(config_uri)
    env = bootstrap(config_uri)
    here = os.path.dirname(__file__)
    sitemap_dir = os.path.join(here, "..", "static", "_sitemaps")
    if os.listdir(sitemap_dir):
        print(
            "[WARNING]The target sitemap directory ({}) is not empty.\n"
            "Existing sitemaps may get overridden. But the sitemap index file "
            "will contain all sitemaps in the folder, even old ones that are "
            "no longer needed. Consider deleting the contents of the folder "
            "first.".format(os.path.abspath(sitemap_dir))
        )
        if not args.no_input:
            input("Press [Enter] to continue.")

    create_sitemaps(settings, limit, sitemap_dir, env)


if __name__ == '__main__':  # pragma: no cover
    main()



"""
Module that sets up the datamanagers and the database connections.
"""
from skosprovider_sqlalchemy.models import Base as SkosBase
from sqlalchemy import engine_from_config
from sqlalchemy.orm import sessionmaker
from zope.sqlalchemy import register

from atramhasis.data.datamanagers import AuditManager
from atramhasis.data.datamanagers import ConceptSchemeManager
from atramhasis.data.datamanagers import LanguagesManager
from atramhasis.data.datamanagers import SkosManager
from .models import Base


def data_managers(request):
    """
    Generate a datamanager with a database session and register a cleanup handler.

    :param pyramid.request.Request request: The request this db session will
        be tied to.
    :returns: A dictionary containing different
        :class:`datamanagers <atramhasis.data.datamanagers.DataManager>`.
    """
    session = request.db
    skos_manager = SkosManager(session)
    conceptscheme_manager = ConceptSchemeManager(session)
    languages_manager = LanguagesManager(session)
    audit_manager = AuditManager(session)

    return {'skos_manager': skos_manager, 'conceptscheme_manager': conceptscheme_manager,
            'languages_manager': languages_manager, 'audit_manager': audit_manager}


def db(request):
    session = request.registry.dbmaker()

    def cleanup(_):
        session.close()
    request.add_finished_callback(cleanup)
    return session


def includeme(config):
    """
    Set up SQLAlchemy.

    :param pyramid.config.Configurator config: Pyramid configuration.
    """

    # Setting up SQLAlchemy
    engine = engine_from_config(config.get_settings(), 'sqlalchemy.')
    Base.metadata.bind = engine
    SkosBase.metadata.bind = engine
    config.registry.dbmaker = sessionmaker(bind=engine)
    register(config.registry.dbmaker)
    config.add_request_method(data_managers, reify=True)
    config.add_request_method(db, reify=True)



import argparse
import csv
import json
import os
import sys

from rdflib import Graph
from rdflib.util import SUFFIX_FORMAT_MAP
from rdflib.util import guess_format
from skosprovider.providers import DictionaryProvider
from skosprovider.providers import SimpleCsvProvider
from skosprovider.skos import ConceptScheme
from skosprovider.uri import UriPatternGenerator
from skosprovider_rdf.providers import RDFProvider
from skosprovider_sqlalchemy.utils import import_provider
from sqlalchemy import create_engine
from sqlalchemy.engine import url
from sqlalchemy.orm import sessionmaker

from atramhasis.data.models import Provider
from atramhasis.scripts.migrate_sqlalchemy_providers import json_serial


def file_to_rdf_provider(**kwargs) -> RDFProvider:
    """
    Create RDF provider from the input file
    """
    input_file = kwargs.get('input_file')
    input_name, input_ext = os.path.splitext(os.path.basename(input_file))
    meta_id = kwargs.get("provider_id") or input_name.upper()
    graph = Graph()
    graph.parse(input_file, format=guess_format(input_ext))
    return RDFProvider(
        {'id': meta_id},
        graph
    )


def _create_provider_kwargs(**kwargs):
    provider_kwargs = {}
    uri_pattern = kwargs.get('uri_pattern')
    if uri_pattern:
        provider_kwargs['uri_generator'] =  UriPatternGenerator(uri_pattern)
    concept_scheme = kwargs.get('concept_scheme')
    if concept_scheme:
        provider_kwargs['concept_scheme'] = concept_scheme
    return provider_kwargs


def file_to_csv_provider(**kwargs) -> SimpleCsvProvider:
    """
    Create CSV provider from the input file
    """
    input_file = kwargs.get('input_file')
    input_name, input_ext = os.path.splitext(os.path.basename(input_file))
    meta_id = kwargs.get("provider_id") or input_name.upper()
    provider_kwargs = _create_provider_kwargs(**kwargs)
    with open(input_file) as ifile:
        reader = csv.reader(ifile)
        return SimpleCsvProvider(
            {'id': meta_id},
            reader,
            **provider_kwargs
        )


def file_to_json_provider(**kwargs) -> DictionaryProvider:
    """
    Create Dictionary provider from the input file
    """
    input_file = kwargs.get('input_file')
    input_name, input_ext = os.path.splitext(os.path.basename(input_file))
    meta_id = kwargs.get("provider_id") or input_name.upper()
    provider_kwargs = _create_provider_kwargs(**kwargs)
    with open(input_file) as data_file:
        dictionary = json.load(data_file)
    return DictionaryProvider(
        {'id': meta_id},
        dictionary,
        **provider_kwargs
    )


supported_types = {
    'JSON': {
        'extensions': ['.json'],
        'file_to_provider': file_to_json_provider
    },
    'RDF': {
        'extensions': ['.%s' % suffix for suffix in SUFFIX_FORMAT_MAP],
        'file_to_provider': file_to_rdf_provider
    },
    'CSV': {
        'extensions': ['.csv'],
        'file_to_provider': file_to_csv_provider
    }
}

supported_ext = [item for sublist in [supported_types[filetype]['extensions'] for filetype in supported_types.keys()]
                 for item in sublist]


def parse_argv_for_import(argv):
    """
    Parse parameters and validate
    """
    cmd = os.path.basename(argv[0])
    parser = argparse.ArgumentParser(
        description='Import file to a database',
        epilog=(
            f'example: {cmd} '
            'atramhasis/scripts/my_file '
            'urn:x-skosprovider:trees:%s '
            '--to sqlite:///atramhasis.sqlite '
            '--conceptscheme-label Labels '
            '--conceptscheme-uri urn:x-skosprovider:trees '
            '--create-provider '
            '--provider-id ERFGOEDTYPES '
            '--id-generation-strategy numeric'
        )
    )
    parser.add_argument(
        'input_file',
        type=str,
        help='local path to the input file',
    )
    parser.add_argument(
        'uri_pattern',
        type=str,
        help='URI pattern input for the URIGenerator',
    )
    parser.add_argument(
        '--to',
        dest='to',
        metavar='conn_string',
        type=str,
        help='Connection string of the output database',
        required=False,
        default='sqlite:///atramhasis.sqlite'
    )
    parser.add_argument(
        '--conceptscheme-label',
        dest='cs_label',
        type=str,
        help='Label of the conceptscheme',
        required=False,
        default=None
    )
    parser.add_argument(
        '--conceptscheme-uri',
        dest='cs_uri',
        type=str,
        help='URI of the conceptscheme',
        required=False,
        default=None
    )
    parser.add_argument(
        '--create-provider',
        dest='create_provider',
        default=True,
        action=argparse.BooleanOptionalAction,
        help='An optional parameter if given a provider is created. '
             'Use --no-create-provider to not create a provider',
    )
    parser.add_argument(
        '--provider-id',
        dest='provider_id',
        type=str,
        help='An optional string (eg. ERFGOEDTYPES) to be assigned to the provider id. '
             'If not specified, assign the conceptscheme id to the provider id',
        required=False,
        default=None
    )
    parser.add_argument(
        '--id-generation-strategy',
        dest='id_generation_strategy',
        type=str,
        help='URI pattern input for the URIGenerator',
        required=False,
        choices=["numeric", "guid", "manual"],
        default="numeric"
    )
    args = parser.parse_args()
    if not validate_file(args.input_file) or not validate_connection_string(args.to):
        sys.exit(1)
    return args


def validate_file(input_file):
    if not os.path.exists(input_file):
        print(f'The input file {input_file} does not exists')
        return False
    elif os.path.splitext(input_file)[1] not in supported_ext:
        print(f'the input file {input_file} is not supported. Allowed extensions are: {supported_ext}')
        return False
    else:
        return True


def validate_connection_string(connection_string):
    """
    Validate the connection string
    :param connection_string
    :return: Boolean True if correct connection string
    """
    u = url.make_url(connection_string)
    if u.drivername == 'postgresql':
        if u.username and u.password and u.host and u.port and u.database:
            return True
    elif u.drivername == 'sqlite':
        if u.database:
            return True
    elif u.drivername:
        print('The database driver ' + u.drivername + ' is not supported.')
    print('Wrong structure of connection string "' + connection_string + '"')
    print('Structure: postgresql://username:password@host:port/db_name OR sqlite:///path/db_name.sqlite')
    return False


def conn_str_to_session(conn_str):
    """
    create session from database connection string
    """
    connect_uri = conn_str
    engine = create_engine(connect_uri, echo=True)
    return sessionmaker(
        bind=engine,
    )()


def create_conceptscheme(conceptscheme_uri: str, conceptscheme_label: str) -> ConceptScheme:
    """
    Create a conceptscheme based on arg values
    """
    return ConceptScheme(
            uri=conceptscheme_uri,
            labels = [{'label': conceptscheme_label}]
        )


def main(argv=sys.argv):
    """
    Documentation: import -h
    Run: import
    <path_input_file>
    <uri_pattern>
    --to <conn_string>
    --conceptscheme-uri <cs_uri>
    --conceptscheme-label <cs_label>
    --create-provider
    --provider-id <provider_id>
    --id-generation-strategy <numeric/guid/manual>

    example path_input_file:
     atramhasis/scripts/my_file

    structure conn_string:
     postgresql://username:password@host:port/db_name
     sqlite:///path/db_name.sqlite
    default conn_string:
     sqlite:///atramhasis.sqlite

    example conceptscheme_label
     My Conceptscheme
    default conceptscheme_label is the name of the file if a URI is specified.
    If no URI is specified, a conceptscheme will be imported from the input
    file. This only works for RDf files. For other file types (JSON and CSV)
    conceptscheme_uri is mandatory and conceptscheme_label is recommended.
    """

    # Import the data
    args = parse_argv_for_import(argv)
    input_name, input_ext = os.path.splitext(os.path.basename(args.input_file))
    session = conn_str_to_session(args.to)
    file_to_provider_function = [
        supported_types[filetype]['file_to_provider']
        for filetype in supported_types.keys()
        if input_ext in supported_types[filetype]['extensions']
    ][0]
    if args.cs_uri:
        cs_uri = args.cs_uri
        cs_label = args.cs_label if args.cs_label else input_name.capitalize()
        args.concept_scheme = create_conceptscheme(cs_uri, cs_label)
    provider = file_to_provider_function(**vars(args))
    cs = import_provider(provider, session)
    if args.create_provider:
        db_provider = Provider()
        provider.metadata[
            'atramhasis.id_generation_strategy'
        ] = args.id_generation_strategy.upper()
        db_provider.meta = json.loads(json.dumps(provider.metadata, default=json_serial))
        db_provider.expand_strategy = 'RECURSE'
        db_provider.conceptscheme = cs
        db_provider.id = args.provider_id or cs.id
        db_provider.uri_pattern = args.uri_pattern
        if 'conceptscheme_id' in db_provider.meta:
            del db_provider.meta['conceptscheme_id']
        session.add(db_provider)
    session.commit()

    # Get info to return to the user
    scheme_id = cs.id
    if not args.create_provider:
        prov_id = getattr(args, 'provider_id', None) or input_name.upper()
        print(
            "\n\n*** The import of conceptscheme {0} from the {1} file to {2} was succesful. ***\
              \n\nTo use the data in Atramhasis, you must edit the file my_thesaurus/skos/__init__.py.\
              \nAdd a configuration similar to:\
                \n\ndef create_registry(request):\
                \n\t# create the SKOS registry\
                \n\tregistry = Registry(instance_scope='threaded_thread')\
                \n\t{3} = SQLAlchemyProvider(\
                \n\t\t{{'id': '{4}', 'conceptscheme_id': {5}}},\
                \n\t\trequest.db\
                \n\t)\
                \n\tregistry.register_provider({6})\
                \n\treturn registry\
                \n\n".
            format(
                prov_id, args.input_file, args.to,
                prov_id.replace(' ', '_'), prov_id, scheme_id, prov_id.replace(' ', '_')
            )
        )
    else:
        prov_id = args.provider_id or cs.id
        msg = """
***
The import of conceptscheme {0} from the {1} file to {2} was succesful.
You can now continue through the Atramhasis UI.
***
"""
        print(msg.format(prov_id, args.input_file, args.to))



if __name__ == '__main__':
    main()



import logging

import requests
from cachecontrol import CacheControl
from cachecontrol.heuristics import ExpiresAfter
from skosprovider.registry import Registry
from skosprovider_getty.providers import AATProvider
from skosprovider_getty.providers import TGNProvider
from sqlalchemy.orm import Session

from atramhasis import utils
from atramhasis.data.datamanagers import ProviderDataManager

log = logging.getLogger(__name__)
LICENSES = [
    'https://creativecommons.org/licenses/by/4.0/',
    'http://data.vlaanderen.be/doc/licentie/modellicentie-gratis-hergebruik/v1.0'
]


def register_providers_from_db(registry: Registry, session: Session) -> None:
    """
    Retrieve all providers stored in the database and add them to the registry.

    :param registry: The registry to which the providers will be added.
    :param session: A database session.
    """
    manager = ProviderDataManager(session)
    for db_provider in manager.get_all_providers():
        provider = utils.db_provider_to_skosprovider(db_provider)
        registry.register_provider(provider)


def create_registry(request):
    try:
        registry = Registry(instance_scope='threaded_thread')

        getty_session = CacheControl(requests.Session(), heuristic=ExpiresAfter(weeks=1))

        aat = AATProvider(
            {'id': 'AAT', 'subject': ['external']},
            session=getty_session
        )

        tgn = TGNProvider(
            {'id': 'TGN', 'subject': ['external']},
            session=getty_session
        )

        registry.register_provider(aat)
        registry.register_provider(tgn)
        register_providers_from_db(registry, request.db)

        return registry
    except AttributeError:
        log.exception("Attribute error during creation of Registry.")
        raise



import unittest
from datetime import date
from unittest.mock import Mock

from rdflib import DCTERMS
from rdflib import FOAF
from rdflib import Graph
from rdflib import RDF
from rdflib import VOID
from rdflib.term import Literal
from rdflib.term import URIRef

from atramhasis.rdf import HYDRA
from atramhasis.rdf import _add_ldf_server
from atramhasis.rdf import _add_metadataset
from atramhasis.rdf import _add_provider


class AddProviderTests(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def _get_duri(self):
        return URIRef('http://test.atramhasis.org/void.ttl#bigdataset')

    def _get_graph(self, duri):
        g = Graph()
        g.add((duri, RDF.type, VOID.Dataset))
        return g

    def _get_provider(self, dataset={}):
        provider_mock = Mock()
        provider_mock.get_vocabulary_id = Mock(return_value='TREES')
        provider_mock.get_metadata = Mock(return_value={'id': 'TREES', 'subject': [], 'dataset': dataset})
        cs = Mock()
        cs.uri = 'urn:x-atramhasis:conceptschemes:trees'
        cs.labels = []
        provider_mock.concept_scheme = cs
        return provider_mock

    def testEmptyProvider(self):
        duri = self._get_duri()
        g = self._get_graph(duri)
        sduri = 'http://test.atramhasis.org/void.ttl#smalldataset'
        p = self._get_provider({'uri': sduri})
        homepage = 'http://test.atramhasis.org/conceptschemes/TREES'
        rdfdump = 'http://test.atramhasis.org/conceptschemes/TREES/c.rdf'
        ttldump = 'http://test.atramhasis.org/conceptschemes/TREES/c.rdf'
        req = Mock()
        req.route_url = Mock()
        req.route_url.side_effect = [
            sduri,
            homepage,
            'http://test.atramhasis.org/conceptschemes/TREES/c.rdf',
            'http://test.atramhasis.org/conceptschemes/TREES/c.ttl'
        ]
        req.registry.settings = {
            'atramhasis.ldf.enabled': True,
            'atramhasis.ldf.baseurl': 'http://test.atramhasis.org/ldf'
        }
        g = _add_provider(g, p, duri, req)
        sd = URIRef(sduri)
        self.assertIn((duri, RDF.type, VOID.Dataset), g)
        self.assertIn((sd, RDF.type, VOID.Dataset), g)
        self.assertIn((duri, VOID.subset, sd), g)
        self.assertIn((sd, DCTERMS.identifier, Literal('TREES')), g)
        self.assertIn((sd, VOID.rootResource, URIRef(p.concept_scheme.uri)), g)
        self.assertIn((sd, FOAF.homepage, URIRef(homepage)), g)
        self.assertIn((sd, VOID.dataDump, URIRef(rdfdump)), g)
        self.assertIn((sd, VOID.dataDump, URIRef(ttldump)), g)
        self.assertIn((sd, HYDRA.search, None), g)


class MetadatasetTests(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def _get_graph(self):
        return Graph()

    def test_add_empty_metadataset(self):
        metadataset = {}
        g = self._get_graph()
        uri = URIRef('http://test.atramhasis.org/void.ttl#emptyset')
        g.add((uri, RDF.type, VOID.Dataset))
        g = _add_metadataset(g, uri, metadataset)
        assert len(g) == 1

    def test_add_metadataset(self):
        metadataset = {
            'publisher': ['https://id.erfgoed.net/actoren/501'],
            'created': [date(year=2016, month=9, day=14)],
            'language': ['nl', 'en', 'fr'],
            'license': [
                'https://creativecommons.org/licenses/by/4.0/',
                'http://data.vlaanderen.be/doc/licentie/modellicentie-gratis-hergebruik/v1.0'
            ]
        }
        g = self._get_graph()
        uri = URIRef('http://id.python.org/datasets/different_types_of_trees')
        g.add((uri, RDF.type, VOID.Dataset))
        g = _add_metadataset(g, uri, metadataset)
        assert len(g) == 8
        self.assertIn((uri, DCTERMS.language, Literal('nl')), g)
        self.assertIn((uri, DCTERMS.language, Literal('fr')), g)
        self.assertIn((uri, DCTERMS.language, Literal('en')), g)
        self.assertIn((uri, DCTERMS.publisher, URIRef('https://id.erfgoed.net/actoren/501')), g)
        self.assertIn((uri, DCTERMS.license, URIRef('https://creativecommons.org/licenses/by/4.0/')), g)
        self.assertIn((uri, DCTERMS.license, URIRef('http://data.vlaanderen.be/doc/licentie/modellicentie-gratis-hergebruik/v1.0')), g)
        self.assertIn((uri, DCTERMS.created, Literal(date(year=2016, month=9, day=14))), g)


class LdfServerTests(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def _get_graph(self):
        return Graph()

    def test_add_ldf_server(self):
        url = Literal('http://demo.atramhasis.org/ldf{?s,p,o}')
        g = self._get_graph()
        uri = URIRef('http://test.atramhasis.org/void.ttl#emptyset')
        g.add((uri, RDF.type, VOID.Dataset))
        g = _add_ldf_server(g, uri, url)
        self.assertIn((uri, HYDRA.search, None), g)
        self.assertIn((None, HYDRA.template, url), g)
        self.assertIn((None, HYDRA.mapping, None), g)



"""note_markup

Revision ID: 3ac8aca026fd
Revises: 1ad2b6fbcf22
Create Date: 2015-08-17 10:46:02.444677

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "3ac8aca026fd"
down_revision = "1ad2b6fbcf22"


def upgrade():
    op.add_column("note", sa.Column("markup", sa.String(20)))


def downgrade():
    with op.batch_alter_table("note") as batch_op:
        batch_op.drop_column("markup")



import unittest
from unittest.mock import Mock

from skosprovider_sqlalchemy.models import Collection
from skosprovider_sqlalchemy.models import Concept
from skosprovider_sqlalchemy.models import ConceptScheme
from skosprovider_sqlalchemy.models import Label
from skosprovider_sqlalchemy.models import Match
from skosprovider_sqlalchemy.models import MatchType
from skosprovider_sqlalchemy.models import Source
from sqlalchemy.exc import NoResultFound

from atramhasis import mappers
from atramhasis.data.models import ExpandStrategy
from atramhasis.data.models import IDGenerationStrategy
from atramhasis.data.models import Provider
from atramhasis.mappers import map_concept
from atramhasis.mappers import map_conceptscheme

test_json = {
    "narrower": [{"id": 8}, {"id": 7}, {"id": 9}],
    "label": "Belgium",
    "type": "concept",
    "id": 4,
    "broader": [{"id": 2}, {"id": 11}],
    "related": [{"id": 5}, {"id": 12}],
    "member_of": [{"id": 999}],
    "labels": [{
                   "label": "Belgium",
                   "type": "prefLabel",
                   "language": "en"
               }, {
                   "label": "België",
                   "type": "prefLabel",
                   "language": "nl"
               }],
    "notes": [{
                  "note": "een notitie",
                  "type": "note",
                  "language": "nl"
              }],
    "sources": [{
        "citation": "Atlas."
    }]
}
json_collection = {
    "id": 0,
    "labels": [{
                   "language": "nl",
                   "label": "Stijlen en culturen",
                   "type": "prefLabel"
               }],
    "type": "collection",
    "label": "Stijlen en culturen",
    "members": [{"id": 61}, {"id": 60}, {"id": 12}],
    "member_of": [{"id": 999}, {"id": 7}],
    "notes": [{
                  "note": "een notitie",
                  "type": "note",
                  "language": "nl"
              }]
}
test_json_conceptscheme = {
    "label": "Stijlen en culturen",
    "labels": [{
                   "language": "nl",
                   "label": "Stijlen en culturen",
                   "type": "prefLabel"
               }],
    "notes": [{
                  "note": "een notitie",
                  "type": "note",
                  "language": "nl"
              }],
    "sources": [{
        "citation": "Atlas."
    }]
}
test_json_html = {
    "narrower": [{"id": 8}, {"id": 7}, {"id": 9}],
    "label": "Belgium",
    "type": "concept",
    "id": 4,
    "broader": [{"id": 2}, {"id": 11}],
    "related": [{"id": 5}, {"id": 12}],
    "member_of": [{"id": 999}],
    "labels": [{
                   "label": "Belgium",
                   "type": "prefLabel",
                   "language": "en"
               }, {
                   "label": "België",
                   "type": "prefLabel",
                   "language": "nl"
               }],
    "notes": [{
                  "note": "een <a href='#'>notitie</a>",
                  "type": "note",
                  "language": "nl"
              }],
    "sources": [{
        "citation": "<em>Atlas.</em>"
    }]
}


class DummySkosManager:
    def get_thing(self, concept_id, conceptscheme_id):
        if concept_id in (7, 11, 12):
            raise NoResultFound()
        if concept_id in (999, 19):
            return Collection(
                id=concept_id,
                concept_id=concept_id,
                conceptscheme_id=conceptscheme_id
            )
        else:
            return Concept(
                id=concept_id,
                concept_id=concept_id,
                conceptscheme_id=conceptscheme_id
            )

    def change_type(self, _, concept_id, conceptscheme_id, new_type, uri):
        thing = Concept() if new_type == 'concept' else Collection()
        thing.type = new_type
        thing.concept_id = concept_id
        thing.conceptscheme_id = conceptscheme_id
        thing.uri = uri
        return thing

    def get_match_type(self, *_, **__):
        return Mock()

    def get_match(self, uri, matchtype_id, concept_id):
        if concept_id == -1:
            raise NoResultFound()
        else:
            match = Match()
            match.uri = uri
            match.concept_id = concept_id
            if matchtype_id in [
                'broadMatch', 'closeMatch', 'exactMatch', 'narrowMatch', 'relatedMatch'
            ]:
                match.matchtype = MatchType(matchtype_id, 'test')
                return match
            else:
                raise NoResultFound()


class TestMappers(unittest.TestCase):
    def setUp(self):
        self.skos_manager = DummySkosManager()
        self.concept = Concept()
        self.concept.concept_id = 1
        self.concept.conceptscheme_id = 1
        self.collection = Collection()
        self.collection.concept_id = 0
        self.collection.conceptscheme_id = 1
        self.collection.uri = "urn:x-skosprovider:trees/3"
        self.conceptscheme = ConceptScheme()
        self.conceptscheme.id = 1
        member_concept_1 = Concept()
        member_concept_1.concept_id = 5
        member_concept_1.conceptscheme_id = 1
        member_concept_2 = Collection()
        member_concept_2.concept_id = 6
        member_concept_2.conceptscheme_id = 1
        self.collection.members.add(member_concept_1)
        self.collection.members.add(member_concept_2)
        self.concept.narrower_concepts.add(member_concept_1)
        self.concept.narrower_collections.add(member_concept_2)

    def tearDown(self):
        self.concept = None

    def test_mapping(self):
        result_concept = map_concept(self.concept, test_json, self.skos_manager)
        self.assertIsNotNone(result_concept)
        self.assertEqual(3, len(result_concept.narrower_concepts))
        self.assertEqual(2, len(result_concept.broader_concepts))
        self.assertEqual(2, len(result_concept.related_concepts))
        self.assertEqual(1, len(result_concept.member_of))
        self.assertEqual(2, len(result_concept.labels))
        self.assertEqual(1, len(result_concept.notes))
        self.assertEqual(1, len(result_concept.sources))
        self.assertFalse(hasattr(result_concept, 'members'))
        self.assertIsNone(result_concept.notes[0].markup)

    def test_mapping_collections_filled(self):
        label = Label(label='test', labeltype_id='altLabel', language_id='nl')
        self.concept.labels.append(label)
        related_concept = Concept(concept_id=6, conceptscheme_id=1)
        self.concept.related_concepts.add(related_concept)
        source = Source(citation='testCitation')
        self.concept.sources.append(source)
        source = Source(citation='AnotherTestCitation')
        self.concept.sources.append(source)
        result_concept = map_concept(self.concept, test_json, self.skos_manager)
        self.assertEqual(3, len(result_concept.narrower_concepts))
        self.assertEqual(2, len(result_concept.broader_concepts))
        self.assertEqual(2, len(result_concept.related_concepts))
        self.assertEqual(1, len(result_concept.member_of))
        self.assertEqual(2, len(result_concept.labels))
        self.assertEqual(1, len(result_concept.notes))
        self.assertEqual(1, len(result_concept.sources))

    def test_mapping_check_db_lookup(self):
        result_concept = map_concept(self.concept, test_json, self.skos_manager)
        for narrower_concept in result_concept.narrower_concepts:
            self.assertIsNotNone(narrower_concept)
            if narrower_concept.concept_id == 7:
                self.assertIsNone(narrower_concept.id)

    def test_mapping_check_db_lookup_member_of(self):
        result_concept = map_concept(self.concept, test_json, self.skos_manager)
        for member_of in result_concept.member_of:
            self.assertIsNotNone(member_of)
            if member_of.concept_id == 7:
                self.assertIsNone(member_of.id)

    def test_mapping_collection(self):
        result_collection = map_concept(self.collection, json_collection, self.skos_manager)
        self.assertIsNotNone(result_collection)
        self.assertEqual(3, len(result_collection.members))
        self.assertEqual(2, len(result_collection.member_of))
        self.assertEqual(1, len(result_collection.labels))
        self.assertEqual(1, len(result_collection.notes))
        self.assertFalse(hasattr(result_collection, 'related_concepts'))

    def test_mapping_matches(self):
        test_json["matches"] = {"exact": ["urn:sample:666"], "broad": ["urn:somewhere:93"]}
        result_concept = map_concept(self.concept, test_json, self.skos_manager)
        self.assertIsNotNone(result_concept)
        self.assertTrue(hasattr(result_concept, 'matches'))
        self.assertEqual(2, len(result_concept.matches))
        self.assertIn(result_concept.matches[0].uri, ["urn:sample:666", "urn:somewhere:93"])

    def test_mapping_matches_new_concept(self):
        test_json["matches"] = {"exact": ["urn:sample:666"], "broad": ["urn:somewhere:93"]}
        del test_json["id"]
        result_concept = map_concept(self.concept, test_json, self.skos_manager)
        self.assertIsNotNone(result_concept)
        self.assertTrue(hasattr(result_concept, 'matches'))
        self.assertEqual(2, len(result_concept.matches))
        self.assertIn(result_concept.matches[0].uri, ["urn:sample:666", "urn:somewhere:93"])

    def test_mapping_subordinate_arrays(self):
        test_json["subordinate_arrays"] = [{"id": 19}]
        result_concept = map_concept(self.concept, test_json, self.skos_manager)
        self.assertIsNotNone(result_concept)
        self.assertTrue(hasattr(result_concept, 'narrower_collections'))
        self.assertEqual(1, len(result_concept.narrower_collections))
        self.assertEqual([c for c in result_concept.narrower_collections][0].concept_id, 19)

    def test_mapping_subordinate_arrays_no_result(self):
        test_json["subordinate_arrays"] = [{"id": 11}]
        result_concept = map_concept(self.concept, test_json, self.skos_manager)
        self.assertIsNotNone(result_concept)
        self.assertTrue(hasattr(result_concept, 'narrower_collections'))
        self.assertEqual(1, len(result_concept.narrower_collections))
        self.assertEqual([c for c in result_concept.narrower_collections][0].concept_id, 11)

    def test_mapping_superordinates(self):
        json_collection["superordinates"] = [{"id": 12}]
        result_collection = map_concept(self.collection, json_collection, self.skos_manager)
        self.assertIsNotNone(result_collection)
        self.assertTrue(hasattr(result_collection, 'broader_concepts'))
        self.assertEqual(1, len(result_collection.broader_concepts))
        self.assertEqual([c for c in result_collection.broader_concepts][0].concept_id, 12)

    def test_mapping_concept_to_collection(self):
        result_collection = map_concept(self.concept, json_collection, self.skos_manager)
        self.assertIsNotNone(result_collection)
        self.assertTrue(hasattr(result_collection, 'members'))
        self.assertFalse(hasattr(result_collection, 'related_concepts'))
        self.assertFalse(hasattr(result_collection, 'narrower_concepts'))
        self.assertFalse(hasattr(result_collection, 'narrower_collections'))

    def test_mapping_collection_to_concept(self):
        result_concept = map_concept(self.collection, test_json, self.skos_manager)
        self.assertIsNotNone(result_concept)
        self.assertFalse(hasattr(result_concept, 'members'))
        self.assertTrue(hasattr(result_concept, 'related_concepts'))
        self.assertTrue(hasattr(result_concept, 'narrower_concepts'))
        self.assertTrue(hasattr(result_concept, 'narrower_collections'))
        self.assertEqual(result_concept.uri, 'urn:x-skosprovider:trees/3')

    def test_mapping_conceptscheme(self):
        result_conceptscheme = map_conceptscheme(self.conceptscheme, test_json_conceptscheme)
        self.assertIsNotNone(result_conceptscheme)
        self.assertEqual(1, len(result_conceptscheme.labels))
        self.assertEqual(1, len(result_conceptscheme.notes))
        self.assertEqual(1, len(result_conceptscheme.sources))

    def test_mapping_html_note(self):
        result_concept = map_concept(self.concept, test_json_html, self.skos_manager)
        self.assertIsNotNone(result_concept)
        self.assertEqual(3, len(result_concept.narrower_concepts))
        self.assertEqual(2, len(result_concept.broader_concepts))
        self.assertEqual(2, len(result_concept.related_concepts))
        self.assertEqual(1, len(result_concept.member_of))
        self.assertEqual(2, len(result_concept.labels))
        self.assertEqual(1, len(result_concept.notes))
        self.assertEqual(1, len(result_concept.sources))
        self.assertFalse(hasattr(result_concept, 'members'))
        self.assertEqual('HTML', result_concept.notes[0].markup)
        self.assertEqual('HTML', result_concept.sources[0].markup)


def test_map_provider_new_provider_full():
    data = {
        'metadata': {'meta': 'meta'},
        'default_language': 'nl',
        'force_display_language': 'nl-force',
        'id_generation_strategy': 'GUID',
        'subject': ['hidden'],
        'uri_pattern': 'uri-pattern',
        'expand_strategy': 'visit',
        'conceptscheme_uri': 'conceptscheme-uri',
        'id': 'p-id'
    }
    result = mappers.map_provider(data)
    assert result.conceptscheme is not None
    assert result.conceptscheme.uri == 'conceptscheme-uri'
    assert result.id == 'p-id'
    assert result.meta == {
        'atramhasis.force_display_language': 'nl-force',
        'atramhasis.id_generation_strategy': 'GUID',
        'default_language': 'nl',
        'meta': 'meta',
        'subject': ['hidden']
    }
    assert result.default_language == 'nl'
    assert result.force_display_language == 'nl-force'
    assert result.id_generation_strategy is IDGenerationStrategy.GUID
    assert result.subject == ['hidden']
    assert result.uri_pattern == 'uri-pattern'
    assert result.expand_strategy is ExpandStrategy.VISIT


def test_map_provider_new_provider_minimal():
    data = {
        'conceptscheme_uri': 'conceptscheme-uri',
        'uri_pattern': 'uri-pattern',
    }
    result = mappers.map_provider(data)
    assert result.conceptscheme is not None
    assert result.conceptscheme.uri == 'conceptscheme-uri'
    assert result.id is None
    assert result.meta == {
        'atramhasis.force_display_language': None,
        'atramhasis.id_generation_strategy': 'NUMERIC',
        'default_language': None,
        'subject': []
    }
    assert result.default_language is None
    assert result.force_display_language is None
    assert result.id_generation_strategy is IDGenerationStrategy.NUMERIC
    assert result.subject == []
    assert result.uri_pattern == 'uri-pattern'
    assert result.expand_strategy is ExpandStrategy.RECURSE


def test_map_provider_existing():
    data = {
        'metadata': {'meta': 'meta'},
        'default_language': 'nl',
        'force_display_language': 'nl-force',
        'id_generation_strategy': 'GUID',
        'subject': ['hidden'],
        'uri_pattern': 'uri-pattern',
        'expand_strategy': 'visit',
        'id': 'p-id'
    }
    existing = Provider(id='exists')
    result = mappers.map_provider(data, existing)
    assert result is existing
    assert result.conceptscheme is None
    assert result.id == 'exists'
    assert result.meta == {
        'atramhasis.force_display_language': 'nl-force',
        'atramhasis.id_generation_strategy': 'GUID',
        'default_language': 'nl',
        'meta': 'meta',
        'subject': ['hidden']
    }
    assert result.default_language == 'nl'
    assert result.force_display_language == 'nl-force'
    assert result.id_generation_strategy is IDGenerationStrategy.GUID
    assert result.subject == ['hidden']
    assert result.uri_pattern == 'uri-pattern'
    assert result.expand_strategy is ExpandStrategy.VISIT



import os

from pyramid.response import FileResponse
from pyramid.view import view_config


class StaticView:
    """
    Views voor aan de root gebonden static files.
    """

    def __init__(self, request):
        self.request = request
        self.here = os.path.dirname(__file__)

    @view_config(route_name='sitemap')
    def sitemaps(self):
        sitemaps = os.path.join(
            self.here, '..', 'static', '_sitemaps', 'sitemap_index.xml'
        )
        return FileResponse(
            sitemaps,
            request=self.request
        )



from unittest.mock import MagicMock
from unittest.mock import Mock
from unittest.mock import patch

import pytest
from skosprovider.registry import Registry
from skosprovider_getty.providers import GettyProvider

from atramhasis.errors import ValidationError
from atramhasis.json_processors import provider


def test_create_provider_with_id():
    session = Mock()
    with patch.object(provider.mappers, 'map_provider') as mapper:
        mapper.return_value = Mock(id='1')
        data = {}
        response = provider.create_provider(data, session, MagicMock())

    mapper.assert_called()
    session.add.assert_called()
    assert response.id != str(response.conceptscheme.id)


def test_create_provider_no_id():
    session = Mock()
    with patch.object(provider.mappers, 'map_provider') as mapper:
        mapper.return_value = Mock(id=None)
        data = {}
        response = provider.create_provider(data, session, MagicMock())

    mapper.assert_called()
    session.add.assert_called()
    assert response.id == str(response.conceptscheme.id)


def test_create_provider_duplicate_uri():
    session = Mock()
    registry = Registry()
    registry.register_provider(GettyProvider({}))
    with patch.object(provider.mappers, 'map_provider') as mapper:
        mapper.return_value = Mock(id=None)
        data = {"conceptscheme_uri": 'http://vocab.getty.edu/aat/'}
        with pytest.raises(ValidationError) as e:
            provider.create_provider(data, session, registry)
        assert e.value.value == 'Provider could not be validated.'
        assert e.value.errors == [
            {"conceptscheme_uri": "Collides with existing provider."}
        ]


def test_update_provider_with_id():
    session = Mock()
    with patch.object(provider.mappers, 'map_provider') as mapper,\
        patch('atramhasis.json_processors.provider.validate_provider_json') as validator:
        mapper.return_value = Mock(id='1')
        data = {}
        provider.update_provider('1', data, session)

    mapper.assert_called()
    validator.assert_called()


def test_delete_provider_with_id():
    session = Mock()
    with patch.object(provider.mappers, 'map_provider') as mapper, \
        patch.object(provider, 'delete_scheme'):
        mapper.return_value = Mock(id='1')
        provider.delete_provider('1', session)

    session.delete.assert_called()



"""
Module that validates incoming JSON.
"""

import copy

import bleach
import colander
from language_tags import tags
from skosprovider_sqlalchemy.models import (
    Language
)
from sqlalchemy.exc import NoResultFound

from atramhasis.errors import ValidationError
from atramhasis.data.models import IDGenerationStrategy


class ForcedString(colander.String):
    """
    Acts just like colander.String but anything it gets will be turned into a string.

    eg. Passing a number 1 will treat it as "1". null/None remains null/None
    """

    def deserialize(self, node, cstruct):
        if (
            not isinstance(cstruct, str)
            and not isinstance(cstruct, bytes)
            and cstruct is not colander.null
        ):
            cstruct = str(cstruct)
        return super().deserialize(node, cstruct)


class Label(colander.MappingSchema):
    label = colander.SchemaNode(
        colander.String()
    )
    type = colander.SchemaNode(
        colander.String()
    )
    language = colander.SchemaNode(
        colander.String()
    )


def html_preparer(value):
    """
    Prepare the value by stripping all html except certain tags.

    :param value: The value to be cleaned. 
    :rtype: str
    """
    try:
        return bleach.clean(value, tags=['strong', 'em', 'a'], strip=True)
    except TypeError:
        # Trying to clean a non-string
        # Ignore for now so it can be caught later on
        return value


class Note(colander.MappingSchema):
    note = colander.SchemaNode(
        colander.String(),
        preparer=html_preparer
    )
    type = colander.SchemaNode(
        colander.String()
    )
    language = colander.SchemaNode(
        colander.String()
    )


class Source(colander.MappingSchema):
    citation = colander.SchemaNode(
        colander.String(),
        preparer=html_preparer
    )


class Labels(colander.SequenceSchema):
    label = Label()


class Notes(colander.SequenceSchema):
    note = Note()


class Sources(colander.SequenceSchema):
    source = Source()


class RelatedConcept(colander.MappingSchema):
    id = colander.SchemaNode(
        ForcedString()
    )


class Concepts(colander.SequenceSchema):
    concept = RelatedConcept()


class MatchList(colander.SequenceSchema):
    match = colander.SchemaNode(
        colander.String(),
        missing=None
    )


class Matches(colander.MappingSchema):
    broad = MatchList(missing=[])
    close = MatchList(missing=[])
    exact = MatchList(missing=[])
    narrow = MatchList(missing=[])
    related = MatchList(missing=[])


class Concept(colander.MappingSchema):
    id = colander.SchemaNode(
        ForcedString(),
        missing=None
    )
    type = colander.SchemaNode(
        colander.String(),
        missing='concept'
    )
    labels = Labels(missing=[])
    notes = Notes(missing=[])
    sources = Sources(missing=[])
    broader = Concepts(missing=[])
    narrower = Concepts(missing=[])
    related = Concepts(missing=[])
    members = Concepts(missing=[])
    member_of = Concepts(missing=[])
    subordinate_arrays = Concepts(missing=[])
    superordinates = Concepts(missing=[])
    matches = Matches(missing={})
    infer_concept_relations = colander.SchemaNode(
        colander.Boolean(),
        missing=colander.drop
    )

    def preparer(self, concept):
        return concept


class ConceptScheme(colander.MappingSchema):
    labels = Labels(missing=[])
    notes = Notes(missing=[])
    sources = Sources(missing=[])


class LanguageTag(colander.MappingSchema):
    id = colander.SchemaNode(
        colander.String()
    )
    name = colander.SchemaNode(
        colander.String()
    )


def concept_schema_validator(node, cstruct):
    """
    This validator validates an incoming concept or collection

    This validator will run a list of rules against the concept or collection
    to see that there are no validation rules being broken.

    :param colander.SchemaNode node: The schema that's being used while validating.
    :param cstruct: The concept or collection being validated.
    """
    request = node.bindings['request']
    validate_id_generation = node.bindings['validate_id_generation']
    skos_manager = request.data_managers['skos_manager']
    languages_manager = request.data_managers['languages_manager']
    provider = node.bindings['provider']
    conceptscheme_id = provider.conceptscheme_id
    concept_type = cstruct['type']
    collection_id = cstruct['id']
    narrower = None
    broader = None
    related = None
    members = None
    member_of = None
    r_validated = False
    n_validated = False
    b_validated = False
    m_validated = False
    o_validated = False
    errors = []
    min_labels_rule(errors, node, cstruct)
    if 'labels' in cstruct:
        labels = copy.deepcopy(cstruct['labels'])
        label_type_rule(errors, node, skos_manager, labels)
        label_lang_rule(errors, node, languages_manager, labels)
        max_preflabels_rule(errors, node, labels)
    if 'related' in cstruct:
        related = copy.deepcopy(cstruct['related'])
        related = [m['id'] for m in related]
        r_validated = semantic_relations_rule(errors, node['related'], skos_manager,
                                              conceptscheme_id, related, collection_id)
        concept_relations_rule(errors, node['related'], related, concept_type)
    if 'narrower' in cstruct:
        narrower = copy.deepcopy(cstruct['narrower'])
        narrower = [m['id'] for m in narrower]
        n_validated = semantic_relations_rule(errors, node['narrower'], skos_manager,
                                              conceptscheme_id, narrower, collection_id)
        concept_relations_rule(errors, node['narrower'], narrower, concept_type)
    if 'broader' in cstruct:
        broader = copy.deepcopy(cstruct['broader'])
        broader = [m['id'] for m in broader]
        b_validated = semantic_relations_rule(errors, node['broader'], skos_manager,
                                              conceptscheme_id, broader, collection_id)
        concept_relations_rule(errors, node['broader'], broader, concept_type)
    if 'members' in cstruct:
        members = copy.deepcopy(cstruct['members'])
        members = [m['id'] for m in members]
        m_validated = semantic_relations_rule(errors, node['members'], skos_manager,
                                              conceptscheme_id, members, collection_id)
    if 'member_of' in cstruct:
        member_of = copy.deepcopy(cstruct['member_of'])
        member_of = [m['id'] for m in member_of]
        o_validated = semantic_relations_rule(errors, node['member_of'], skos_manager,
                                              conceptscheme_id, member_of, collection_id)
    if r_validated and n_validated and b_validated:
        concept_type_rule(errors, node['narrower'], skos_manager, conceptscheme_id, narrower)
        narrower_hierarchy_rule(errors, node['narrower'], skos_manager, conceptscheme_id, cstruct)
        concept_type_rule(errors, node['broader'], skos_manager, conceptscheme_id, broader)
        broader_hierarchy_rule(errors, node['broader'], skos_manager, conceptscheme_id, cstruct)
        concept_type_rule(errors, node['related'], skos_manager, conceptscheme_id, related)

    if m_validated and o_validated:
        members_only_in_collection_rule(errors, node['members'], concept_type, members)
        collection_members_unique_rule(errors, node['members'], members)
        collection_type_rule(errors, node['member_of'], skos_manager, conceptscheme_id, member_of)
        memberof_hierarchy_rule(errors, node['member_of'], skos_manager, conceptscheme_id, cstruct)
        members_hierarchy_rule(errors, node['members'], skos_manager, conceptscheme_id, cstruct)

    if 'matches' in cstruct:
        matches = copy.deepcopy(cstruct['matches'])
        concept_matches_rule(errors, node['matches'], matches, concept_type)
        concept_matches_unique_rule(errors, node['matches'], matches)

    if 'subordinate_arrays' in cstruct:
        subordinate_arrays = copy.deepcopy(cstruct['subordinate_arrays'])
        subordinate_arrays = [m['id'] for m in subordinate_arrays]
        subordinate_arrays_only_in_concept_rule(errors, node['subordinate_arrays'], concept_type, subordinate_arrays)
        subordinate_arrays_type_rule(errors, node['subordinate_arrays'], skos_manager, conceptscheme_id,
                                     subordinate_arrays)
        subordinate_arrays_hierarchy_rule(errors, node['subordinate_arrays'], skos_manager, conceptscheme_id, cstruct)

    if 'superordinates' in cstruct:
        superordinates = copy.deepcopy(cstruct['superordinates'])
        superordinates = [m['id'] for m in superordinates]
        superordinates_only_in_concept_rule(errors, node['superordinates'], concept_type, superordinates)
        superordinates_type_rule(errors, node['superordinates'], skos_manager, conceptscheme_id, superordinates)
        superordinates_hierarchy_rule(errors, node['superordinates'], skos_manager, conceptscheme_id, cstruct)

    if cstruct['type'] == 'concept' and 'infer_concept_relations' in cstruct:
        msg = "'infer_concept_relations' can only be set for collections."
        errors.append(colander.Invalid(node['infer_concept_relations'], msg=msg))

    if validate_id_generation:
        id_generation_strategy = provider.metadata.get(
            "atramhasis.id_generation_strategy", IDGenerationStrategy.NUMERIC
        )
        if id_generation_strategy == IDGenerationStrategy.MANUAL:
            if not cstruct.get("id"):
                msg = "Required for this provider."
                errors.append(colander.Invalid(node["id"], msg=msg))
            else:
                try:
                    skos_manager.get_thing(cstruct["id"], conceptscheme_id)
                except NoResultFound:
                    # this is desired
                    pass
                else:
                    msg = f"{cstruct['id']} already exists."
                    errors.append(colander.Invalid(node["id"], msg=msg))

    if len(errors) > 0:
        raise ValidationError(
            'Concept could not be validated',
            [e.asdict() for e in errors]
        )


def conceptscheme_schema_validator(node, cstruct):
    """
    This validator validates the incoming conceptscheme labels

    :param colander.SchemaNode node: The schema that's being used while validating.
    :param cstruct: The conceptscheme being validated.
    """
    request = node.bindings['request']
    skos_manager = request.data_managers['skos_manager']
    languages_manager = request.data_managers['languages_manager']
    errors = []
    min_labels_rule(errors, node, cstruct)
    if 'labels' in cstruct:
        labels = copy.deepcopy(cstruct['labels'])
        label_type_rule(errors, node, skos_manager, labels)
        label_lang_rule(errors, node, languages_manager, labels)
        max_preflabels_rule(errors, node, labels)
    if len(errors) > 0:
        raise ValidationError(
            'ConceptScheme could not be validated',
            [e.asdict() for e in errors]
        )


def concept_relations_rule(errors, node_location, relations, concept_type):
    """
    Checks that only concepts have narrower, broader and related relations.
    """
    if relations is not None and len(relations) > 0 and concept_type != 'concept':
        errors.append(colander.Invalid(
            node_location,
            'Only concepts can have narrower/broader/related relations'
        ))


def max_preflabels_rule(errors, node, labels):
    """
    Checks that there's only one prefLabel for a certain language.
    """
    preflabel_found = []
    for label in labels:
        if label['type'] == 'prefLabel':
            if label['language'] in preflabel_found:
                errors.append(colander.Invalid(
                    node['labels'],
                    'Only one prefLabel per language allowed.'
                ))
            else:
                preflabel_found.append(label['language'])


def min_labels_rule(errors, node, cstruct):
    """
    Checks that a label or collection always has a least one label.
    """
    if 'labels' in cstruct:
        labels = copy.deepcopy(cstruct['labels'])
        if len(labels) == 0:
            errors.append(colander.Invalid(
                node['labels'],
                'At least one label is necessary'
            ))


def label_type_rule(errors, node, skos_manager, labels):
    """
    Checks that a label has the correct type.
    """
    label_types = skos_manager.get_all_label_types()
    label_types = [label_type.name for label_type in label_types]
    for label in labels:
        if label['type'] not in label_types:
            errors.append(colander.Invalid(
                node['labels'],
                'Invalid labeltype.'
            ))


def label_lang_rule(errors, node, languages_manager, labels):
    """
    Checks that languages of a label are valid.

    Checks that they are valid IANA language tags. If the language tag was not
    already present in the database, it adds them.
    """
    for label in labels:
        language_tag = label['language']
        if not tags.check(language_tag):
            errors.append(colander.Invalid(
                node['labels'],
                'Invalid language tag: %s' % ", ".join([err.message for err in tags.tag(language_tag).errors])
            ))
        else:
            languages_present = languages_manager.count_languages(language_tag)
            if not languages_present:
                descriptions = ', '.join(tags.description(language_tag))
                language_item = Language(id=language_tag, name=descriptions)
                languages_manager.save(language_item)


def concept_type_rule(errors, node_location, skos_manager, conceptscheme_id, items):
    """
    Checks that the targets of narrower, broader and related are concepts and
    not collections.
    """
    for item_concept_id in items:
        item_concept = skos_manager.get_thing(item_concept_id, conceptscheme_id)
        if item_concept.type != 'concept':
            errors.append(colander.Invalid(
                node_location,
                'A narrower, broader or related concept should always be a concept, not a collection'
            ))


def collection_type_rule(errors, node_location, skos_manager, conceptscheme_id, members):
    """
    Checks that the targets of member_of are collections and not concepts.
    """
    for member_collection_id in members:
        member_collection = skos_manager.get_thing(member_collection_id, conceptscheme_id)
        if member_collection.type != 'collection':
            errors.append(colander.Invalid(
                node_location,
                'A member_of parent should always be a collection'
            ))


def semantic_relations_rule(errors, node_location, skos_manager, conceptscheme_id, members, collection_id):
    """
    Checks that the elements in a group of concepts or collections are not the
    the group itself, that they actually exist and are within
    the same conceptscheme.
    """
    for member_concept_id in members:
        if member_concept_id == collection_id:
            errors.append(colander.Invalid(
                node_location,
                'A concept or collection cannot be related to itself'
            ))
            return False
        try:
            skos_manager.get_thing(member_concept_id, conceptscheme_id)
        except NoResultFound:
            errors.append(colander.Invalid(
                node_location,
                'Concept not found, check concept_id. Please be aware members should be within one scheme'
            ))
            return False
    return True


def hierarchy_build(skos_manager, conceptscheme_id, property_list, property_hierarchy, property_concept_type,
                    property_list_name):
    for property_concept_id in property_list:
        try:
            property_concept = skos_manager.get_thing(property_concept_id, conceptscheme_id)
        except NoResultFound:
            property_concept = None
        if property_concept is not None and (
                        property_concept.type == property_concept_type or property_concept_type is None):
            property_concepts = [n.concept_id for n in getattr(property_concept, property_list_name)]
            for members_id in property_concepts:
                property_hierarchy.append(members_id)
                hierarchy_build(skos_manager, conceptscheme_id, property_concepts, property_hierarchy,
                                property_concept_type, property_list_name)


def hierarchy_rule(errors, node_location, skos_manager, conceptscheme_id, cstruct, property1, property2,
                   property2_list_name, concept_type, error_message):
    """
    Checks that the property1 of a concept are not already in property2 hierarchy

    """
    property2_hierarchy = []
    property1_list = []
    if property1 in cstruct:
        property1_value = copy.deepcopy(cstruct[property1])
        property1_list = [m['id'] for m in property1_value]
    if property2 in cstruct:
        property2_value = copy.deepcopy(cstruct[property2])
        property2_list = [m['id'] for m in property2_value]
        property2_hierarchy = property2_list
        hierarchy_build(skos_manager, conceptscheme_id, property2_list, property2_hierarchy, concept_type,
                        property2_list_name)
    for broader_concept_id in property1_list:
        if broader_concept_id in property2_hierarchy:
            errors.append(colander.Invalid(
                node_location,
                error_message
            ))


def broader_hierarchy_rule(errors, node_location, skos_manager, conceptscheme_id, cstruct):
    """
    Checks that the broader concepts of a concepts are not alreadt narrower
    concepts of that concept.
    """
    hierarchy_rule(errors, node_location, skos_manager, conceptscheme_id, cstruct, 'broader', 'narrower',
                   'narrower_concepts', 'concept',
                   'The broader concept of a concept must not itself be a narrower concept of the concept being edited.'
                   )


def narrower_hierarchy_rule(errors, node_location, skos_manager, conceptscheme_id, cstruct):
    """
    Checks that the narrower concepts of a concept are not already broader
    concepts of that concept.
    """
    hierarchy_rule(errors, node_location, skos_manager, conceptscheme_id, cstruct, 'narrower', 'broader',
                   'broader_concepts', 'concept',
                   'The narrower concept of a concept must not itself be a broader concept of the concept being edited.'
                   )


def collection_members_unique_rule(errors, node_location, members):
    """
    Checks that a collection has no duplicate members.
    """
    if len(members) > len(set(members)):
        errors.append(colander.Invalid(
            node_location,
            'All members of a collection should be unique.'
        ))


def members_only_in_collection_rule(errors, node, concept_type, members):
    """
    Checks that only collections have members.
    """
    if concept_type != 'collection' and len(members) > 0:
        errors.append(colander.Invalid(
            node,
            'Only collections can have members.'
        ))


def memberof_hierarchy_rule(errors, node_location, skos_manager, conceptscheme_id, cstruct):
    hierarchy_rule(errors, node_location, skos_manager, conceptscheme_id, cstruct, 'member_of', 'members',
                   'members', 'collection',
                   'The parent member_of collection of a concept must not itself be a member of the concept being edited.'
                   )


def members_hierarchy_rule(errors, node_location, skos_manager, conceptscheme_id, cstruct):
    """
    Checks that a collection does not have members that are in themselves
    already "parents" of that collection.
    """
    hierarchy_rule(errors, node_location, skos_manager, conceptscheme_id, cstruct, 'members', 'member_of',
                   'member_of', 'collection',
                   'The item of a members collection must not itself be a parent of the concept/collection being edited.'
                   )


def concept_matches_rule(errors, node_location, matches, concept_type):
    """
    Checks that only concepts have matches.
    """
    if matches is not None and len(matches) > 0 and concept_type != 'concept':
        errors.append(colander.Invalid(
            node_location,
            'Only concepts can have matches'
        ))


def concept_matches_unique_rule(errors, node_location, matches):
    """
    Checks that a concept has not duplicate matches.

    This means that a concept can only have one match (no matter what the type)
    with another concept. We don't allow eg. a concept that has both a broadMatch
    and a relatedMatch with the same concept.
    """
    if matches is not None:
        uri_list = []
        for matchtype in matches:
            uri_list.extend([uri for uri in matches[matchtype]])
        if len(uri_list) > len(set(uri_list)):
            errors.append(colander.Invalid(
                node_location,
                'All matches of a concept should be unique.'
            ))


def languagetag_validator(node, cstruct):
    """
    This validator validates a languagetag.

    The validator will check if a tag is a valid IANA language tag. The the
    validator is informed that this should be a new language tag, it will also
    check if the tag doesn't already exist.

    :param colander.SchemaNode node: The schema that's being used while validating.
    :param cstruct: The value being validated.
    """
    request = node.bindings['request']
    languages_manager = request.data_managers['languages_manager']
    new = node.bindings['new']
    errors = []
    language_tag = cstruct['id']

    if new:
        languagetag_checkduplicate(node['id'], language_tag, languages_manager, errors)
    languagetag_isvalid_rule(node['id'], language_tag, errors)

    if len(errors) > 0:
        raise ValidationError(
            'Language could not be validated',
            [e.asdict() for e in errors]
        )


def languagetag_isvalid_rule(node, language_tag, errors):
    """
    Check that a languagetag is a valid IANA language tag.
    """
    if not tags.check(language_tag):
        errors.append(colander.Invalid(
            node,
            'Invalid language tag: %s' % ", ".join([err.message for err in tags.tag(language_tag).errors])
        ))


def languagetag_checkduplicate(node, language_tag, languages_manager, errors):
    """
    Check that a languagetag isn't duplicated.
    """
    language_present = languages_manager.count_languages(language_tag)
    if language_present:
        errors.append(colander.Invalid(
            node,
            'Duplicate language tag: %s' % language_tag)
        )


def subordinate_arrays_only_in_concept_rule(errors, node, concept_type, subordinate_arrays):
    """
    Checks that only a concept has subordinate arrays.
    """
    if concept_type != 'concept' and len(subordinate_arrays) > 0:
        errors.append(colander.Invalid(
            node,
            'Only concept can have subordinate arrays.'
        ))


def subordinate_arrays_type_rule(errors, node_location, skos_manager, conceptscheme_id, subordinate_arrays):
    """
    Checks that subordinate arrays are always collections.
    """
    for subordinate_id in subordinate_arrays:
        subordinate = skos_manager.get_thing(subordinate_id, conceptscheme_id)
        if subordinate.type != 'collection':
            errors.append(colander.Invalid(
                node_location,
                'A subordinate array should always be a collection'
            ))


def subordinate_arrays_hierarchy_rule(errors, node_location, skos_manager, conceptscheme_id, cstruct):
    """
    Checks that the subordinate arrays of a concept are not themselves
    parents of that concept.
    """
    hierarchy_rule(errors, node_location, skos_manager, conceptscheme_id, cstruct, 'subordinate_arrays', 'member_of',
                   'members', 'collection',
                   'The subordinate_array collection of a concept must not itself be a parent of the concept being edited.'
                   )


def superordinates_only_in_concept_rule(errors, node, concept_type, superordinates):
    """
    Checks that only collections have superordinates.
    """
    if concept_type != 'collection' and len(superordinates) > 0:
        errors.append(colander.Invalid(
            node,
            'Only collection can have superordinates.'
        ))


def superordinates_type_rule(errors, node_location, skos_manager, conceptscheme_id, superordinates):
    """
    Checks that superordinates are always concepts.
    """
    for superordinate_id in superordinates:
        superordinate = skos_manager.get_thing(superordinate_id, conceptscheme_id)
        if superordinate.type != 'concept':
            errors.append(colander.Invalid(
                node_location,
                'A superordinate should always be a concept'
            ))


def superordinates_hierarchy_rule(errors, node_location, skos_manager, conceptscheme_id, cstruct):
    """
    Checks that the superordinate concepts of a collection are not themselves
    members of that collection.
    """
    hierarchy_rule(errors, node_location, skos_manager, conceptscheme_id, cstruct, 'superordinates', 'members',
                   'members', 'collection',
                   'The superordinates of a collection must not itself be a member of the collection being edited.'
                   )


def validate_provider_json(json_data, provider_id=None):
    errors = []

    if provider_id:
        if provider_id != json_data.get('id'):
            errors.append(
                {"id": "Id does not match with id parameter in url."},
            )
    # Do not allow keys in the metadata which exist in the root of the json.
    forbidden_metadata_keys = (
        'default_language',
        'subject',
        'force_display_language',  # while not the same, this would just be confusing
        'atramhasis.force_display_language',
        'id_generation_strategy',  # while not the same, this would just be confusing
        'atramhasis.id_generation_strategy',
    )
    metadata = json_data.get('metadata', {})
    if wrong_keys := [k for k in forbidden_metadata_keys if k in metadata]:
        errors.append({'metadata': f'Found disallowed key(s): {", ".join(wrong_keys)}.'})

    if json_data.get('default_language'):
        if not tags.check(json_data['default_language']):
            errors.append({'default_language': 'Invalid language.'})

    if json_data.get('force_display_language'):
        if not tags.check(json_data['force_display_language']):
            errors.append({'force_display_language': 'Invalid language.'})

    if json_data.get('subject'):
        if json_data['subject'] != ['hidden']:
            errors.append({'subject': 'Subject must be one of: "hidden"'})

    if errors:
        raise ValidationError('Provider could not be validated.', errors)



import os
import random
import shutil

from pyramid.paster import get_appsettings

from atramhasis import main
from tests import DbTest
from tests import setup_db

here = os.path.dirname(__file__)
settings = get_appsettings(os.path.join(here, '../', 'tests/conf_test.ini'))


def setUpModule():
    setup_db()


class TestConfig(DbTest):

    def test_config(self):
        app = main({}, **settings)
        self.assertIsNotNone(app)

    def test_config_alt_dump_location(self):
        random_dump_loc = f'dump{random.randint(0, 100)}'
        settings['atramhasis.dump_location'] = os.path.join(here, random_dump_loc)
        app = main({}, **settings)
        self.assertIsNotNone(app)
        if os.path.exists(random_dump_loc):
            shutil.rmtree(random_dump_loc)



"""
Routes for the Atramhasis views.

.. versionadded:: 0.4.4
"""
import os


def includeme(config):
    """
    Setup the routing for Atramhasis.

    :param pyramid.config.Configurator config: The application config.
    """

    config.add_rewrite_rule(r'/(?P<path>.*)/', r'/%(path)s')
    config.add_route('home', '/')

    # Configure pyramid_openapi3 integration
    config.pyramid_openapi3_spec(
        os.path.join(os.path.dirname(__file__), "openapi.yaml"),
        route="/api_doc/openapi.yaml",
    )
    config.pyramid_openapi3_add_explorer(route="/api_doc")

    config.add_static_view('sitemaps', 'static/_sitemaps/', cache_max_age=3600)
    config.add_static_view('static', 'static', cache_max_age=3600)
    config.add_route("sitemap", "/sitemap_index.xml")

    config.add_route('atramhasis.rdf_void_turtle_ext', pattern='/void.ttl', accept='text/turtle')
    config.add_route('atramhasis.rdf_full_export_ext', pattern='/conceptschemes/{scheme_id}/c.rdf')
    config.add_route('atramhasis.rdf_full_export_turtle_ext', pattern='/conceptschemes/{scheme_id}/c.ttl')
    config.add_route('atramhasis.rdf_conceptscheme_export_ext', pattern='/conceptschemes/{scheme_id}.rdf')
    config.add_route('atramhasis.rdf_conceptscheme_export_turtle_ext', pattern='/conceptschemes/{scheme_id}.ttl')
    config.add_route('atramhasis.rdf_individual_export_ext', pattern='/conceptschemes/{scheme_id}/c/{c_id:.*}.rdf')
    config.add_route('atramhasis.rdf_individual_export_turtle_ext', pattern='/conceptschemes/{scheme_id}/c/{c_id:.*}.ttl')
    config.add_route('atramhasis.rdf_conceptscheme_jsonld_ext', pattern='/conceptschemes/{scheme_id}.jsonld')
    config.add_route('atramhasis.rdf_individual_jsonld_ext', pattern='/conceptschemes/{scheme_id}/c/{c_id:.*}.jsonld')

    config.add_route('conceptschemes', pattern='/conceptschemes', accept='text/html', request_method="GET")
    config.add_route('conceptscheme', pattern='/conceptschemes/{scheme_id}', accept='text/html', request_method="GET")
    config.add_route('concept', pattern='/conceptschemes/{scheme_id}/c/{c_id:.*}', accept='text/html',
                     request_method="GET")
    config.add_route('search_result', pattern='/conceptschemes/{scheme_id}/c', accept='text/html')
    config.add_route('scheme_root', pattern='/conceptschemes/{scheme_id}/c/', accept='text/html')
    config.add_route('scheme_tree_html', pattern='/conceptschemes/{scheme_id}/tree', accept='text/html')
    config.add_route('scheme_tree', pattern='/conceptschemes/{scheme_id}/tree', accept='application/json')

    config.add_route('search_result_export', pattern='/conceptschemes/{scheme_id}/c.csv')
    config.add_route('atramhasis.edit_conceptscheme', pattern='/conceptschemes/{scheme_id}',
                     accept='application/json', request_method='PUT')
    config.add_route('atramhasis.get_conceptscheme', pattern='/conceptschemes/{scheme_id}', accept='application/json')
    config.add_route('atramhasis.get_conceptschemes', pattern='/conceptschemes', accept='application/json')
    config.add_route('atramhasis.get_concept', pattern='/conceptschemes/{scheme_id}/c/{c_id:.*}',
                     accept='application/json', request_method="GET")
    config.add_route('atramhasis.add_concept', pattern='/conceptschemes/{scheme_id}/c', accept='application/json',
                     request_method="POST")
    config.add_route('atramhasis.edit_concept', pattern='/conceptschemes/{scheme_id}/c/{c_id:.*}',
                     accept='application/json', request_method="PUT")
    config.add_route('atramhasis.delete_concept', pattern='/conceptschemes/{scheme_id}/c/{c_id:.*}',
                     accept='application/json', request_method="DELETE")
    config.add_route('atramhasis.list_languages', pattern='/languages', accept='application/json',
                     request_method="GET")
    config.add_route('atramhasis.get_language', pattern='/languages/{l_id}', accept='application/json',
                     request_method="GET")
    config.add_route('atramhasis.edit_language', pattern='/languages/{l_id}', accept='application/json',
                     request_method="PUT")
    config.add_route('atramhasis.delete_language', pattern='/languages/{l_id}', accept='application/json',
                     request_method="DELETE")
    config.add_route('locale', '/locale')
    config.add_route('labeltypes', '/labeltypes', accept='application/json', request_method="GET")
    config.add_route('notetypes', '/notetypes', accept='application/json', request_method="GET")

    config.add_route('admin', '/admin')
    config.add_route('scheme_tree_invalidate', pattern='/admin/tree/invalidate/{scheme_id}', accept='application/json')
    config.add_route('tree_invalidate', pattern='/admin/tree/invalidate', accept='application/json')

    config.add_route('atramhasis.rdf_full_export_turtle', pattern='/conceptschemes/{scheme_id}/c', accept='text/turtle')
    config.add_route('atramhasis.rdf_full_export_turtle_x', pattern='/conceptschemes/{scheme_id}/c',
                     accept='application/x-turtle')
    config.add_route('atramhasis.rdf_full_export', pattern='/conceptschemes/{scheme_id}/c',
                     accept='application/rdf+xml')
    config.add_route('atramhasis.rdf_conceptscheme_export', pattern='/conceptschemes/{scheme_id}',
                     accept='application/rdf+xml')
    config.add_route('atramhasis.rdf_conceptscheme_export_turtle', pattern='/conceptschemes/{scheme_id}',
                     accept='text/turtle')
    config.add_route('atramhasis.rdf_conceptscheme_export_turtle_x', pattern='/conceptschemes/{scheme_id}',
                     accept='application/x-turtle')
    config.add_route('atramhasis.rdf_individual_export', pattern='/conceptschemes/{scheme_id}/c/{c_id:.*}',
                     accept='application/rdf+xml')
    config.add_route('atramhasis.rdf_individual_export_turtle', pattern='/conceptschemes/{scheme_id}/c/{c_id:.*}',
                     accept='text/turtle')
    config.add_route('atramhasis.rdf_individual_export_turtle_x', pattern='/conceptschemes/{scheme_id}/c/{c_id:.*}',
                     accept='application/x-turtle')
    config.add_route('atramhasis.rdf_conceptscheme_jsonld', pattern='/conceptschemes/{scheme_id}',
                     accept='application/ld+json')
    config.add_route('atramhasis.rdf_individual_jsonld', pattern='/conceptschemes/{scheme_id}/c/{c_id:.*}',
                     accept='application/ld+json')
    config.add_route('atramhasis.providers', pattern='/providers')
    config.add_route('atramhasis.provider', pattern='/providers/{id}')



from datetime import date
from datetime import datetime
from unittest.mock import Mock
from unittest.mock import patch

from skosprovider_sqlalchemy.models import Concept
from skosprovider_sqlalchemy.models import ConceptScheme
from skosprovider_sqlalchemy.models import LabelType
from skosprovider_sqlalchemy.models import Language
from sqlalchemy import select
from sqlalchemy.exc import NoResultFound

from atramhasis.data.datamanagers import AuditManager
from atramhasis.data.datamanagers import ConceptSchemeManager
from atramhasis.data.datamanagers import CountsManager
from atramhasis.data.datamanagers import LanguagesManager
from atramhasis.data.datamanagers import ProviderDataManager
from atramhasis.data.datamanagers import SkosManager
from atramhasis.data.models import ConceptVisitLog
from atramhasis.data.models import ConceptschemeCounts
from atramhasis.data.models import IDGenerationStrategy
from atramhasis.data.models import Provider
from tests import DbTest
from tests import fill_db
from tests import setup_db


def setUpModule():
    setup_db()
    fill_db()


class ConceptSchemeManagerTest(DbTest):
    def setUp(self):
        super().setUp()
        self.conceptscheme_manager = ConceptSchemeManager(self.session)

    def test_get(self):
        res = self.conceptscheme_manager.get(1)
        self.assertEqual('urn:x-skosprovider:trees', res.uri)

    def test_find(self):
        query = {'type': 'concept', 'label': 'es'}
        res = self.conceptscheme_manager.find(1, query)
        self.assertEqual(1, len(res))

    def test_get_concepts_for_scheme_tree(self):
        res = self.conceptscheme_manager.get_concepts_for_scheme_tree(2)
        self.assertEqual(1, len(res))

    def test_get_collections_for_scheme_tree(self):
        res = self.conceptscheme_manager.get_collections_for_scheme_tree(2)
        self.assertEqual(1, len(res))

    def test_get_all(self):
        res = self.conceptscheme_manager.get_all(2)
        self.assertEqual(10, len(res))

    def test_save(self):
        conceptscheme = self.session.execute(
            select(ConceptScheme).filter(Concept.id == 1)
        ).scalars().first()
        conceptscheme = self.conceptscheme_manager.save(conceptscheme)
        self.assertIsNotNone(conceptscheme.id)


class SkosManagerTest(DbTest):
    def setUp(self):
        super().setUp()
        self.skos_manager = SkosManager(self.session)

    def test_get_thing(self):
        res = self.skos_manager.get_thing(1, 1)
        self.assertEqual('urn:x-skosprovider:trees/1', res.uri)

    def test_save(self):
        thing = Concept()
        thing.concept_id = 123
        thing.conceptscheme_id = 1
        thing = self.skos_manager.save(thing)
        self.assertIsNotNone(thing.id)

    def test_delete_thing(self):
        thing = self.skos_manager.get_thing(1, 1)
        self.skos_manager.delete_thing(thing)
        self.assertRaises(NoResultFound, self.skos_manager.get_thing, 1, 1)

    def test_get_by_list_type(self):
        res = self.skos_manager.get_by_list_type(LabelType)
        self.assertEqual(4, len(res))

    def test_get_match_type(self):
        match_type = self.skos_manager.get_match_type('narrowMatch')
        self.assertEqual('narrowMatch', match_type.name)

    def test_get_match(self):
        from skosprovider_sqlalchemy.models import Match
        match = Match()
        match.matchtype_id = 'narrowMatch'
        match.uri = 'urn:test'
        match.concept_id = 1
        self.session.add(match)
        match = self.skos_manager.get_match('urn:test', 'narrowMatch', 1)
        self.assertEqual('urn:test', match.uri)

    def test_get_all_label_types(self):
        res = self.skos_manager.get_all_label_types()
        self.assertEqual(4, len(res))

    def test_get_next_cid_numeric(self):
        res = self.skos_manager.get_next_cid(1, IDGenerationStrategy.NUMERIC)
        self.assertIsInstance(res, int)

    def test_get_next_cid_guid(self):
        res = self.skos_manager.get_next_cid(1, IDGenerationStrategy.GUID)
        self.assertIsInstance(res, str)
        char = "[0-9a-fA-F]"
        self.assertRegex(
            res,
            fr"^{char}{{8}}\b-{char}{{4}}\b-{char}{{4}}\b-{char}{{4}}\b-{char}{{12}}$"
        )

    def test_get_next_cid_manual(self):
        with self.assertRaises(ValueError):
            self.skos_manager.get_next_cid(1, IDGenerationStrategy.MANUAL)


class LanguagesManagerTest(DbTest):
    def setUp(self):
        super().setUp()
        self.language_manager = LanguagesManager(self.session)

    def test_get(self):
        res = self.language_manager.get('nl')
        self.assertEqual('Dutch', res.name)

    def test_save(self):
        language = Language('au', 'Austrian')
        language = self.language_manager.save(language)
        self.assertEqual('Austrian', language.name)

    def test_delete(self):
        language = self.language_manager.get('en')
        self.language_manager.delete(language)
        self.assertRaises(NoResultFound, self.language_manager.get, 'en')

    def test_get_all(self):
        res = self.language_manager.get_all()
        self.assertGreaterEqual(len(res), 3)

    def test_get_all_sorted(self):
        result = self.language_manager.get_all_sorted('id', False)
        result_ids = [lang.id for lang in result]
        self.assertEqual(list(sorted(result_ids)), result_ids)

    def test_get_all_sorted_desc(self):
        result = self.language_manager.get_all_sorted('id', True)
        result_ids = [lang.id for lang in result]
        self.assertEqual(list(sorted(result_ids, reverse=True)), result_ids)

    def test_count_languages(self):
        res = self.language_manager.count_languages('nl')
        self.assertEqual(1, res)


class AuditManagerTest(DbTest):
    def setUp(self):
        super().setUp()
        self.audit_manager = AuditManager(self.session)

    @patch('atramhasis.data.datamanagers.date', Mock(today=Mock(return_value=date(2015, 8, 1))))
    def test_get_first_day(self):
        self.assertEqual('2015-07-31', self.audit_manager._get_first_day('last_day'))
        self.assertEqual('2015-07-25', self.audit_manager._get_first_day('last_week'))
        self.assertEqual('2015-07-01', self.audit_manager._get_first_day('last_month'))
        self.assertEqual('2014-08-01', self.audit_manager._get_first_day('last_year'))

    @patch('atramhasis.data.datamanagers.date', Mock(today=Mock(return_value=date(2015, 9, 15))))
    def test_get_most_popular_concepts_for_conceptscheme(self):
        self.session.add(
            ConceptVisitLog(concept_id=1, conceptscheme_id='1', origin='REST',
                            visited_at=datetime(2015, 8, 27, 10, 58, 3))
        )
        self.session.add(
            ConceptVisitLog(concept_id=1, conceptscheme_id='1', origin='REST',
                            visited_at=datetime(2015, 8, 27, 11, 58, 3))
        )
        self.session.add(
            ConceptVisitLog(concept_id=2, conceptscheme_id='1', origin='REST',
                            visited_at=datetime(2015, 8, 27, 10, 58, 3))
        )
        self.session.add(
            ConceptVisitLog(concept_id=2, conceptscheme_id='2', origin='REST',
                            visited_at=datetime(2015, 8, 27, 10, 58, 3))
        )
        self.assertListEqual(
            [
                {'concept_id': 1, 'scheme_id': 1},
                {'concept_id': 2, 'scheme_id': 1}
            ],
            self.audit_manager.get_most_popular_concepts_for_conceptscheme(
                1, 5, 'last_month'
            )
        )
        self.assertListEqual([{'concept_id': 2, 'scheme_id': 2}],
                             self.audit_manager.get_most_popular_concepts_for_conceptscheme(2, 5, 'last_month'))
        self.assertListEqual([{'concept_id': 1, 'scheme_id': 1}],
                             self.audit_manager.get_most_popular_concepts_for_conceptscheme(1, 1, 'last_month'))
        self.assertListEqual([],
                             self.audit_manager.get_most_popular_concepts_for_conceptscheme(1, 5, 'last_day'))


class CountsManagerTest(DbTest):
    def setUp(self):
        super().setUp()
        self.counts_manager = CountsManager(self.session)

    def test_count_for_scheme(self):
        counts = ConceptschemeCounts()
        counts.conceptscheme_id = 'TREES'
        counts.counted_at = datetime.now()
        counts.triples = 3
        counts.conceptscheme_triples = 2
        counts.avg_concept_triples = 1
        self.counts_manager.save(counts)
        res = self.counts_manager.get_most_recent_count_for_scheme('TREES')
        self.assertIsNotNone(res)
        self.assertEqual(3, res.triples)


class ProviderDataManagerTest(DbTest):
    def setUp(self):
        super().setUp()
        self.manager = ProviderDataManager(self.session)

    def test_get_provider_by_id(self):
        provider = Provider(
            id='a', conceptscheme=ConceptScheme(), uri_pattern='u-p', meta={}
        )
        self.session.add(provider)
        self.session.flush()

        result = self.manager.get_provider_by_id(provider.id)
        self.assertEqual(result, provider)

    def test_get_provider_by_id_no_result(self):
        with self.assertRaises(NoResultFound):
            self.manager.get_provider_by_id('...')

    def test_get_all_providers(self):
        provider = Provider(
            id='a', conceptscheme=ConceptScheme(), uri_pattern='u-p', meta={}
        )
        self.session.add(provider)
        self.session.flush()

        result = self.manager.get_all_providers()
        self.assertEqual(result, [provider])



import logging
import os
import sys
from unittest.mock import Mock
from unittest.mock import patch

import skosprovider.skos
from pyramid.paster import get_appsettings
from pyramid.request import Request
from skosprovider.exceptions import ProviderUnavailableException
from skosprovider.providers import DictionaryProvider
from skosprovider_sqlalchemy.models import Concept
from skosprovider_sqlalchemy.models import ConceptScheme
from sqlalchemy.orm import sessionmaker
from webtest import TestApp

from atramhasis import main
from atramhasis.cache import list_region
from atramhasis.cache import tree_region
from atramhasis.data.models import Provider
from atramhasis.protected_resources import ProtectedResourceEvent
from atramhasis.protected_resources import ProtectedResourceException
from fixtures.data import chestnut
from fixtures.data import larch
from fixtures.data import species
from tests import DbTest
from tests import SETTINGS
from tests import fill_db
from tests import setup_db


def setUpModule():
    setup_db()
    fill_db()


logging.basicConfig(level=logging.DEBUG, stream=sys.stdout)

here = os.path.dirname(__file__)
settings = get_appsettings(os.path.join(here, '../', 'tests/conf_test.ini'))

json_value = {
    "type": "concept",
    "broader": [],
    "narrower": [],
    "related": [],
    "labels": [
        {
            "type": "prefLabel",
            "language": "en",
            "label": "The Larch"
        },
        {
            "type": "sortLabel",
            "language": "en",
            "label": "a"
        }
    ],
    "notes": [],
    "sources": [
        {
            "citation": "Python, M.: Episode Three: How to recognise different types of trees from quite a long way away."
        }
    ]
}

json_value_relations = {
    "broader": [{"id": 12}],
    "id": 13,
    "related": [],
    "type": "concept",
    "labels": [{
        "label": "koperlegeringen",
        "language": "nl",
        "type": "prefLabel"
    }],
    "label": "koperlegeringen",
    "notes": [],
    "narrower": [{"id": 15}, {"id": 14}]
}

json_value_invalid = """{
    "type": "concept",
    "broader": [],
    "narrower": [],
    "related"[]: [],
    "labels": [
        {
            "type": "prefLabel",
            "language": "en",
            "label": "The Larch"
        }
    ],
    "notes": []}
}"""

json_collection_value = {
    "labels": [{
        "language": "nl",
        "label": "Test verzameling",
        "type": "prefLabel"
    }],
    "type": "collection",
    "label": "Test verzameling",
    "members": [{"id": 333}, {"id": 7}],
    "notes": [{
        "note": "een notitie",
        "type": "note",
        "language": "nl"
    }],
    'infer_concept_relations': True
}

TEST = DictionaryProvider(
    {
        'id': 'TEST',
        'default_language': 'nl',
        'subject': ['biology']
    },
    [larch, chestnut, species],
    concept_scheme=skosprovider.skos.ConceptScheme('http://id.trees.org')
)


class FunctionalTests(DbTest):
    app = None
    testapp = None

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.init_app()

    @classmethod
    def init_app(cls):
        cls.app = main({}, **SETTINGS)
        cls.testapp = TestApp(cls.app)

        # Commit at end of every request. This will trigger listeners.
        class CommittingRequest(Request):

            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.add_finished_callback(lambda req: req.db.commit())

        cls.testapp.app.request_factory = CommittingRequest
        # change db maker to make sessions bound to the test transaction
        registry = cls.testapp.app.registry
        registry.dbmaker = sessionmaker(bind=cls.connection)

    def setUp(self):
        super().setUp()
        self.testapp.reset()

    @staticmethod
    def mock_event_handler(event):
        if event.uri == 'urn:x-vioe:geography:9':
            referenced_in = ['urn:someobject', 'http://test.test.org/object/2']
            raise ProtectedResourceException(
                f'resource {event.uri} is still in use, preventing operation',
                referenced_in)

    @staticmethod
    def mock_event_handler_provider_unavailable(event):
        if event.uri == 'urn:x-vioe:geography:55':
            raise ProviderUnavailableException('test msg')


class HtmlFunctionalTests(FunctionalTests):
    def _get_default_headers(self):
        return {'Accept': 'text/html'}

    def test_get_home(self):
        res = self.testapp.get('/', headers=self._get_default_headers())
        self.assertEqual('200 OK', res.status)
        self.assertIn('text/html', res.headers['Content-Type'])


class CsvFunctionalTests(FunctionalTests):
    def test_get_csv(self):
        response = self.testapp.get('/conceptschemes/TREES/c.csv?type=collection&label=')
        self.assertEqual('200 OK', response.status)
        self.assertIn('text/csv', response.headers['Content-Type'])
        self.assertIn('attachment;filename="atramhasis_export.csv"',
                      response.headers['Content-Disposition'])

    def test_unicode_csv(self):
        response = self.testapp.get(
            '/conceptschemes/TREES/c.csv?label=Chestnut&_LOCALE_=fr')
        data = response.body.decode('utf-8')
        self.assertIsInstance(data, str)
        self.assertEqual('200 OK', response.status)
        self.assertIn('text/csv', response.headers['Content-Type'])
        self.assertIn('attachment;filename="atramhasis_export.csv"',
                      response.headers['Content-Disposition'])
        self.assertIn('la châtaigne', data)

    def test_get_csv_all(self):
        response = self.testapp.get('/conceptschemes/TREES/c.csv')
        self.assertEqual('200 OK', response.status)
        self.assertIn('text/csv', response.headers['Content-Type'])
        self.assertIn('attachment;filename="atramhasis_export.csv"',
                      response.headers['Content-Disposition'])


class RestFunctionalTests(FunctionalTests):
    def _get_default_headers(self):
        return {'Accept': 'application/json'}

    def test_get_concept(self):
        res = self.testapp.get('/conceptschemes/TREES/c/1',
                               headers=self._get_default_headers())
        self.assertEqual('200 OK', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res.json['id'])
        self.assertEqual(res.json['id'], '1')
        self.assertEqual(res.json['type'], 'concept')
        self.assertIn('sortLabel', [label['type'] for label in res.json['labels']])

    def test_get_conceptscheme(self):
        res = self.testapp.get('/conceptschemes/TREES',
                               headers=self._get_default_headers())
        self.assertEqual('200 OK', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res.json['id'])

    def test_get_concept_dictprovider(self):
        res = self.testapp.get('/conceptschemes/TEST/c/1',
                               headers=self._get_default_headers())
        self.assertEqual('200 OK', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res.json['id'])
        self.assertEqual(res.json['type'], 'concept')

    def test_get_concept_not_found(self):
        res = self.testapp.get('/conceptschemes/TREES/c/89',
                               headers=self._get_default_headers(), status=404,
                               expect_errors=True)
        self.assertEqual('404 Not Found', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])

    def test_get_concept_dictprovider_not_found(self):
        res = self.testapp.get('/conceptschemes/TEST/c/89',
                               headers=self._get_default_headers(), status=404,
                               expect_errors=True)
        self.assertEqual('404 Not Found', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])

    def test_add_concept(self):
        res = self.testapp.post_json('/conceptschemes/TREES/c',
                                     headers=self._get_default_headers(),
                                     params=json_value)
        self.assertEqual('201 Created', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res.json['id'])
        self.assertEqual(res.json['type'], 'concept')

    def test_add_update_concept_manual_id(self):
        json_value['id'] = 'manual-3'
        json_value['sources'][0]['citation'] = 'short'
        res = self.testapp.post_json(
            '/conceptschemes/manual-ids/c',
            headers=self._get_default_headers(),
            params=json_value,
        )
        self.assertEqual(201, res.status_code)
        res_json = res.json
        self.assertDictEqual(
            {
                'id': 'manual-3',
                'type': 'concept',
                'uri': 'urn:x-skosprovider:manual-ids:manual-3',
                'label': 'The Larch',
                'concept_scheme': {
                    'uri': 'urn:x-vioe:manual', 'labels': []
                },
                'labels': [
                    {'label': 'The Larch', 'type': 'prefLabel', 'language': 'en'},
                    {'label': 'a', 'type': 'sortLabel', 'language': 'en'}
                ],
                'notes': [],
                'sources': [
                    {'citation': 'short', 'markup': None}
                ],
                'narrower': [],
                'broader': [],
                'related': [],
                'member_of': [],
                'subordinate_arrays': [],
                'matches': {
                    'close': [], 'exact': [], 'related': [], 'broad': [], 'narrow': []
                }
            },
            res_json
        )
        res_json['labels'][0]['label'] = 'updated'
        res = self.testapp.put_json(
            '/conceptschemes/manual-ids/c/manual-3',
            headers=self._get_default_headers(),
            params=res_json,
        )
        self.assertEqual(200, res.status_code)
        self.assertEqual('updated', res.json['label'])

    def test_add_concept_empty_conceptscheme(self):
        res = self.testapp.post_json('/conceptschemes/STYLES/c',
                                     headers=self._get_default_headers(),
                                     params=json_value)
        self.assertEqual('201 Created', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res.json['id'])

    def test_add_concept_invalid_json(self):
        res = self.testapp.post_json(
            '/conceptschemes/TREES/c', headers=self._get_default_headers(),
            params=json_value_invalid, status=400)
        self.assertEqual('400 Bad Request', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])

    def test_add_concept_conceptscheme_not_found(self):
        res = self.testapp.post_json(
            '/conceptschemes/GARDENNNN/c', headers=self._get_default_headers(),
            params=json_value, status=404,
            expect_errors=True)
        self.assertEqual('404 Not Found', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])

    def test_edit_conceptscheme(self):
        res = self.testapp.put_json(
            '/conceptschemes/TREES', headers=self._get_default_headers(),
            params=json_collection_value)
        self.assertEqual('200 OK', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])

    def test_edit_conceptscheme_invalid(self):
        json_collection_value.pop('labels')
        res = self.testapp.put_json(
            '/conceptschemes/TREES', headers=self._get_default_headers(),
            params=json_collection_value,
            expect_errors=True)
        self.assertEqual('400 Bad Request', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res.json)
        self.assertEqual(res.json, {
            "errors": [{'labels': 'At least one label is necessary'}],
            "message": 'ConceptScheme could not be validated'})
        json_collection_value['labels'] = [{
            "language": "nl",
            "label": "Test verzameling",
            "type": "prefLabel"
        }]

    def test_edit_concept(self):
        res = self.testapp.put_json(
            '/conceptschemes/TREES/c/1', headers=self._get_default_headers(),
            params=json_value)
        self.assertEqual('200 OK', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])

    def test_edit_concept_has_relations(self):
        res = self.testapp.put_json(
            '/conceptschemes/MATERIALS/c/13', headers=self._get_default_headers(),
            params=json_value_relations)
        self.assertEqual('200 OK', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertEqual(2, len(res.json['narrower']))

    def test_edit_concept_not_found(self):
        res = self.testapp.put_json(
            '/conceptschemes/TREES/c/89', headers=self._get_default_headers(),
            params=json_value, status=404,
            expect_errors=True)
        self.assertEqual('404 Not Found', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])

    def test_delete_concept(self):
        new_id = '1'
        res = self.testapp.delete(f'/conceptschemes/TREES/c/{new_id}',
                                  headers=self._get_default_headers())
        self.assertEqual('200 OK', res.status)
        self.assertIsNotNone(res.json['id'])
        self.assertEqual(new_id, res.json['id'])
        from skosprovider_sqlalchemy.models import Concept
        concepten = self.session.query(Concept).all()
        print()

    def test_delete_concept_not_found(self):
        res = self.testapp.delete('/conceptschemes/TREES/c/7895',
                                  headers=self._get_default_headers(),
                                  expect_errors=True)
        self.assertEqual('404 Not Found', res.status)

    def test_add_collection(self):
        res = self.testapp.post_json('/conceptschemes/GEOGRAPHY/c',
                                     headers=self._get_default_headers(),
                                     params=json_collection_value, expect_errors=True)
        self.assertEqual('201 Created', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res.json['id'])
        self.assertEqual(res.json['type'], 'collection')

    def test_edit_collection(self):
        json_collection_value['members'] = [{"id": 7}, {"id": 8}]
        json_collection_value['infer_concept_relations'] = False
        res = self.testapp.put_json('/conceptschemes/GEOGRAPHY/c/333',
                                    headers=self._get_default_headers(),
                                    params=json_collection_value)
        self.assertEqual('200 OK', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res.json['id'])
        self.assertEqual(res.json['type'], 'collection')
        self.assertEqual(2, len(res.json['members']))
        self.assertFalse(res.json['infer_concept_relations'])

    def test_delete_collection(self):
        res = self.testapp.delete('/conceptschemes/GEOGRAPHY/c/333',
                                  headers=self._get_default_headers())
        self.assertEqual('200 OK', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])

    def test_uri(self):
        res = self.testapp.post_json('/conceptschemes/MATERIALS/c',
                                     headers=self._get_default_headers(),
                                     params=json_value)
        self.assertEqual('201 Created', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertEqual('urn:x-vioe:materials:51', res.json['uri'])

    def test_provider_unavailable_view(self):
        def raise_provider_unavailable_exception():
            raise ProviderUnavailableException('test msg')

        with patch('atramhasis.views.crud.AtramhasisCrud.delete_concept',
                   Mock(side_effect=raise_provider_unavailable_exception)):
            res = self.testapp.delete('/conceptschemes/GEOGRAPHY/c/55',
                                      headers=self._get_default_headers(),
                                      expect_errors=True)
            self.assertEqual('503 Service Unavailable', res.status)
            self.assertIn("test msg", res)

    def test_get_languages(self):
        res = self.testapp.get('/languages', headers=self._get_default_headers())
        self.assertEqual('200 OK', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res)
        self.assertEqual(len(res.json), 8)

    def test_get_languages_sort(self):
        res = self.testapp.get('/languages', headers=self._get_default_headers(),
                               params={"sort": "id"})
        self.assertEqual('200 OK', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res)
        self.assertEqual(len(res.json), 8)

    def test_get_languages_sort_desc(self):
        res = self.testapp.get('/languages', headers=self._get_default_headers(),
                               params={"sort": "-id"})
        self.assertEqual('200 OK', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res)
        self.assertEqual(len(res.json), 8)

    def test_get_language(self):
        res = self.testapp.get('/languages/de', headers=self._get_default_headers())
        self.assertEqual('200 OK', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res.json['id'])
        self.assertEqual('German', res.json['name'])

    def test_get_language_not_found(self):
        res = self.testapp.get('/languages/jos', headers=self._get_default_headers(),
                               expect_errors=True)
        self.assertEqual('404 Not Found', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res.json)
        self.assertEqual(res.json, {"message": "The resource could not be found."})

    def test_add_language(self):
        res = self.testapp.put_json('/languages/af', headers=self._get_default_headers(),
                                    params={"id": "af", "name": "Afrikaans"})
        self.assertEqual('200 OK', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res.json['id'])
        self.assertEqual(res.json['name'], 'Afrikaans')

    def test_add_language_non_valid(self):
        res = self.testapp.put_json('/languages/flup',
                                    headers=self._get_default_headers(),
                                    params={"id": "flup", "name": "flup"},
                                    expect_errors=True)
        self.assertEqual('400 Bad Request', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res.json)
        self.assertEqual(res.json, {
            "errors": [{
                "id": "Invalid language tag: Unknown code 'flup', Missing language tag in 'flup'."}],
            "message": "Language could not be validated"})

    def test_add_language_non_valid_json(self):
        res = self.testapp.put_json('/languages/af', headers=self._get_default_headers(),
                                    params={"test": "flup"}, expect_errors=True)
        self.assertEqual('400 Bad Request', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res.json)
        self.assertEqual(res.json, {'errors': {'name': 'Required'},
                                    'message': 'Language could not be validated'})

    def test_edit_language(self):
        res = self.testapp.put_json('/languages/de', headers=self._get_default_headers(),
                                    params={"id": "de", "name": "Duits"})
        self.assertEqual('200 OK', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res.json['id'])
        self.assertEqual(res.json['name'], 'Duits')

    def test_edit_language_invalid_language_tag(self):
        res = self.testapp.put_json('/languages/joss',
                                    headers=self._get_default_headers(),
                                    params={"id": "joss", "name": "Duits"},
                                    expect_errors=True)
        self.assertEqual('400 Bad Request', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res.json)
        self.assertEqual(res.json, {
            'errors': [{
                'id': "Invalid language tag: Unknown code 'joss', Missing language tag in 'joss'."}]
            , "message": "Language could not be validated"})

    def test_edit_language_no_id(self):
        res = self.testapp.put_json('/languages/de', headers=self._get_default_headers(),
                                    params={"name": "Duits"})
        self.assertEqual('200 OK', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res.json['id'])
        self.assertEqual(res.json['name'], 'Duits')

    def test_delete_language(self):
        res = self.testapp.delete('/languages/de', headers=self._get_default_headers())
        self.assertEqual('200 OK', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])

    def test_delete_language_not_found(self):
        res = self.testapp.delete('/languages/jos', headers=self._get_default_headers(),
                                  expect_errors=True)
        self.assertEqual('404 Not Found', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res.json)
        self.assertEqual(res.json, {"message": "The resource could not be found."})

    def test_delete_protected_resource(self):
        def mock_event_handler(event):
            if isinstance(event, ProtectedResourceEvent):
                referenced_in = ['urn:someobject', 'http://test.test.org/object/2']
                raise ProtectedResourceException(
                    'resource {} is still in use, preventing operation'
                    .format(event.uri),
                    referenced_in
                )

        registry = self.testapp.app.registry
        with patch.object(registry, 'notify',
                          new=Mock(side_effect=mock_event_handler)):
            res = self.testapp.delete('/conceptschemes/GEOGRAPHY/c/9',
                                      headers=self._get_default_headers(),
                                      expect_errors=True)
        self.assertEqual('409 Conflict', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res.json)
        self.assertEqual(res.json, {
            "message": "resource urn:x-vioe:geography:9 is still in use, preventing operation",
            "referenced_in": ["urn:someobject", "http://test.test.org/object/2"]
        })

    def test_method_not_allowed(self):
        self.testapp.delete('/conceptschemes/TREES', headers=self._get_default_headers(),
                            status=405)
        self.testapp.post('/conceptschemes', headers=self._get_default_headers(),
                          status=405)

    def test_get_conceptschemes(self):
        self.testapp.get('/conceptschemes', headers=self._get_default_headers(),
                         status=200)

    def test_create_provider_openapi_validation(self):
        response = self.testapp.post_json(
            url='/providers',
            params={
                'uri_pattern': 'invalid',
                'subject': 'wrong'
            },
            headers=self._get_default_headers(),
            expect_errors=True
        )
        self.assertEqual(
            {
                'errors': ['None: Failed to cast value to array type: wrong'],
                'message': 'Request was not valid for schema.'
            },
            response.json
        )
        response = self.testapp.post_json(
            url='/providers',
            params={
                'uri_pattern': 'invalid',
                'subject': ['right']
            },
            headers=self._get_default_headers(),
            expect_errors=True
        )
        self.assertEqual(
            {
                'errors': [
                    "uri_pattern: 'invalid' does not match '.*%s.*'",
                    "uri_pattern: 'conceptscheme_uri' is a required property"
                ],
                'message': 'Request was not valid for schema.'
            },
            response.json
        )

    def test_create_minimal_provider(self):
        response = self.testapp.post_json(
            url='/providers',
            params={
                'conceptscheme_uri': 'https://id.erfgoed.net/thesauri/conceptschemes',
                'uri_pattern': 'https://id.erfgoed.net/thesauri/erfgoedtypes/%s'
            },
            headers=self._get_default_headers(),
            status=201
        )
        self.assertEqual(
            {
                'id': response.json["id"],
                'type': 'SQLAlchemyProvider',
                'conceptscheme_uri': 'https://id.erfgoed.net/thesauri/conceptschemes',
                'uri_pattern': 'https://id.erfgoed.net/thesauri/erfgoedtypes/%s',
                'default_language': None,
                'subject': [],
                'force_display_language': None,
                'metadata': {},
                'id_generation_strategy': 'NUMERIC',
                'expand_strategy': 'recurse'
            },
            response.json
        )

    def test_create_full_provider(self):
        response = self.testapp.post_json(
            url='/providers',
            params={
                'id': 'ERFGOEDTYPES',
                'conceptscheme_uri': 'https://id.erfgoed.net/thesauri/conceptschemes',
                'uri_pattern': 'https://id.erfgoed.net/thesauri/erfgoedtypes/%s',
                'default_language': 'NL',
                'force_display_language': 'NL',
                'subject': ['hidden'],
                'metadata': {'Info': 'Extra data about this provider'},
                'id_generation_strategy': 'MANUAL',
                'expand_strategy': 'visit',
            },
            headers=self._get_default_headers(),
            status=201
        )
        self.assertEqual(
            {
                'id': 'ERFGOEDTYPES',
                'type': 'SQLAlchemyProvider',
                'conceptscheme_uri': 'https://id.erfgoed.net/thesauri/conceptschemes',
                'uri_pattern': 'https://id.erfgoed.net/thesauri/erfgoedtypes/%s',
                'default_language': 'NL',
                'force_display_language': 'NL',
                'subject': ['hidden'],
                'metadata': {'Info': 'Extra data about this provider'},
                'id_generation_strategy': 'MANUAL',
                'expand_strategy': 'visit',
            },
            response.json
        )

    def test_create_full_provider_via_put(self):
        response = self.testapp.put_json(
            url='/providers/ERFGOEDTYPES',
            params={
                'id': 'ERFGOEDTYPES',
                'conceptscheme_uri': 'https://id.erfgoed.net/thesauri/conceptschemes',
                'uri_pattern': 'https://id.erfgoed.net/thesauri/erfgoedtypes/%s',
                'default_language': 'NL',
                'force_display_language': 'NL',
                'subject': ['hidden'],
                'metadata': {'Info': 'Extra data about this provider'},
                'id_generation_strategy': 'MANUAL',
                'expand_strategy': 'visit',
            },
            headers=self._get_default_headers(),
            status=201
        )
        self.assertEqual(
            {
                'id': 'ERFGOEDTYPES',
                'type': 'SQLAlchemyProvider',
                'conceptscheme_uri': 'https://id.erfgoed.net/thesauri/conceptschemes',
                'uri_pattern': 'https://id.erfgoed.net/thesauri/erfgoedtypes/%s',
                'default_language': 'NL',
                'force_display_language': 'NL',
                'subject': ['hidden'],
                'metadata': {'Info': 'Extra data about this provider'},
                'id_generation_strategy': 'MANUAL',
                'expand_strategy': 'visit',
            },
            response.json
        )

    def test_update_provider(self):
        conceptscheme = ConceptScheme(
            uri='https://id.erfgoed.net/thesauri/conceptschemes')
        provider = Provider(
            id='ERFGOEDTYPES',
            uri_pattern='https://id.erfgoed.net/thesauri/erfgoedtypes/%s',
            conceptscheme=conceptscheme,
            meta={},
        )
        self.session.add(provider)
        self.session.flush()

        response = self.testapp.put_json(
            url='/providers/ERFGOEDTYPES',
            params={
                'id': 'ERFGOEDTYPES',
                'type': 'SQLAlchemyProvider',
                'conceptscheme_uri': 'https://id.erfgoed.net/thesauri/conceptschemes',
                'uri_pattern': 'https://id.erfgoed.net/thesauri/updated/%s',
                'default_language': 'NL',
                'subject': ['hidden'],
                'force_display_language': 'NL',
                'metadata': {'extra': 'test-extra'},
                'id_generation_strategy': 'MANUAL',
                'expand_strategy': 'visit'
            },
            headers=self._get_default_headers(),
            status=200
        )

        self.assertEqual(
            {
                'id': 'ERFGOEDTYPES',
                'type': 'SQLAlchemyProvider',
                'conceptscheme_uri': 'https://id.erfgoed.net/thesauri/conceptschemes',
                'uri_pattern': 'https://id.erfgoed.net/thesauri/updated/%s',
                'default_language': 'NL',
                'subject': ['hidden'],
                'force_display_language': 'NL',
                'metadata': {'extra': 'test-extra'},
                'id_generation_strategy': 'MANUAL',
                'expand_strategy': 'visit'
            },
            response.json
        )

    def test_delete_provider_with_concepts(self):
        conceptscheme = ConceptScheme(
            uri='https://id.erfgoed.net/thesauri/conceptschemes')
        concept = Concept(
            concept_id="testconceptje",
            conceptscheme=conceptscheme
        )
        provider = Provider(
            id='ERFGOEDTYPES',
            uri_pattern='https://id.erfgoed.net/thesauri/erfgoedtypes/%s',
            conceptscheme=conceptscheme,
            meta={},
        )
        self.session.add(concept)
        self.session.add(provider)
        self.session.flush()
        conceptscheme_id = conceptscheme.id

        self.session.expire_all()
        self.assertIsNotNone(self.session.get(Provider, 'ERFGOEDTYPES'))

        self.testapp.delete(
            url='/providers/ERFGOEDTYPES',
            headers=self._get_default_headers(),
            status=204
        )
        self.session.expire_all()
        self.assertIsNone(self.session.get(Provider, 'ERFGOEDTYPES'))
        self.assertIsNone(self.session.get(ConceptScheme, conceptscheme_id))

    def test_get_providers(self):
        response = self.testapp.get(
            url='/providers',
            headers=self._get_default_headers(),
            status=200
        )
        self.assertEqual(7, len(response.json))
        response = self.testapp.get(
            url='/providers?subject=biology',
            headers=self._get_default_headers(),
            status=200
        )
        self.assertEqual(
            [
                {
                    'conceptscheme_uri': 'http://id.trees.org',
                    'default_language': 'nl',
                    'force_display_language': None,
                    'id': 'TEST',
                    'subject': ['biology'],
                    'type': 'DictionaryProvider',
                    'uri_pattern': 'urn:x-skosprovider:%s:%s',
                    'metadata': {},
                }
            ],
            response.json)

    def test_get_provider(self):
        response = self.testapp.get(
            url='/providers/GEOGRAPHY',
            headers=self._get_default_headers(),
            status=200
        )
        self.assertEqual(
            {
                'id': 'GEOGRAPHY',
                'type': 'SQLAlchemyProvider',
                'conceptscheme_uri': 'urn:x-vioe:geography',
                'uri_pattern': 'urn:x-vioe:geography:%s',
                'default_language': None,
                'subject': [],
                'force_display_language': None,
                'id_generation_strategy': 'NUMERIC',
                'metadata': {},
                'expand_strategy': 'recurse'
            },
            response.json
        )


class TestCookieView(FunctionalTests):
    def _get_default_headers(self):
        return {'Accept': 'text/html'}

    def test_cookie(self):
        response = self.testapp.get('/locale?language=nl',
                                    headers=self._get_default_headers())
        self.assertIsNotNone(response.headers['Set-Cookie'])
        self.assertEqual(response.status, '302 Found')
        self.assertTrue((response.headers.get('Set-Cookie')).startswith('_LOCALE_=nl'))

    def test_unsupported_language(self):
        config_default_lang = settings.get('pyramid.default_locale_name')
        response = self.testapp.get('/locale?language=fr',
                                    headers=self._get_default_headers())
        self.assertTrue((response.headers.get('Set-Cookie')).startswith(
            '_LOCALE_=' + config_default_lang))


class JsonTreeFunctionalTests(FunctionalTests):
    def _get_default_headers(self):
        return {'Accept': 'application/json'}

    def test_tree(self):
        response = self.testapp.get('/conceptschemes/GEOGRAPHY/tree?_LOCALE_=nl',
                                    headers=self._get_default_headers())
        self.assertEqual('200 OK', response.status)
        self.assertIn('application/json', response.headers['Content-Type'])
        self.assertIsNotNone(response.json)
        self.assertEqual(2, len(response.json))
        self.assertEqual('World', response.json[0]['label'])

    def test_missing_labels(self):
        response = self.testapp.get('/conceptschemes/MISSING_LABEL/tree?_LOCALE_=nl',
                                    headers=self._get_default_headers())
        self.assertEqual('200 OK', response.status)
        self.assertIsNotNone(response.json)
        self.assertEqual(2, len(response.json))
        self.assertEqual('label', response.json[0]['label'])
        self.assertEqual(None, response.json[1]['label'])

    def test_tree_language(self):
        response = self.testapp.get('/conceptschemes/TREES/tree?language=nl',
                                    headers=self._get_default_headers())
        self.assertEqual(200, response.status_code)
        self.assertEqual(
            ['De Lariks', 'De Paardekastanje'],
            [child['label'] for child in response.json[0]['children']]
        )
        response = self.testapp.get('/conceptschemes/TREES/tree?language=en',
                                    headers=self._get_default_headers())
        self.assertEqual(200, response.status_code)
        self.assertEqual(
            ['The Chestnut', 'The Larch'],
            [child['label'] for child in response.json[0]['children']]
        )

    def test_no_tree(self):
        response = self.testapp.get('/conceptschemes/FOO/tree?_LOCALE_=nl',
                                    headers=self._get_default_headers(),
                                    status=404, expect_errors=True)
        self.assertEqual('404 Not Found', response.status)


class HtmlTreeFunctionalTests(FunctionalTests):
    def _get_default_headers(self):
        return {'Accept': 'text/html'}

    def test_tree(self):
        response = self.testapp.get('/conceptschemes/GEOGRAPHY/tree?_LOCALE_=nl',
                                    headers=self._get_default_headers())
        self.assertEqual('200 OK', response.status)
        self.assertIn('text/html', response.headers['Content-Type'])

    def test_no_tree(self):
        response = self.testapp.get('/conceptschemes/FOO/tree?_LOCALE_=nl',
                                    headers=self._get_default_headers(),
                                    status=404, expect_errors=True)
        self.assertEqual('404 Not Found', response.status)


class SkosFunctionalTests(FunctionalTests):

    def _get_default_headers(self):
        return {'Accept': 'text/html'}

    def _get_json_headers(self):
        return {'Accept': 'application/json'}

    def test_admin_no_skos_provider(self):
        with patch.dict(self.app.request_extensions.descriptors):
            del self.app.request_extensions.descriptors['skos_registry']
            res = self.testapp.get('/admin', headers=self._get_default_headers(),
                                   expect_errors=True)
        self.assertEqual('500 Internal Server Error', res.status)
        self.assertTrue('message' in res)
        self.assertTrue(
            'No SKOS registry found, please check your application setup' in res)

    def test_crud_no_skos_provider(self):
        with patch.dict(self.app.request_extensions.descriptors):
            del self.app.request_extensions.descriptors['skos_registry']
            res = self.testapp.post_json('/conceptschemes/GEOGRAPHY/c',
                                         headers=self._get_json_headers(),
                                         params=json_collection_value, expect_errors=True)
        self.assertEqual('500 Internal Server Error', res.status)
        self.assertTrue('message' in res)
        self.assertTrue(
            'No SKOS registry found, please check your application setup' in res)

    def test_match_filter(self):
        response = self.testapp.get(
            '/conceptschemes/TREES/c',
            headers={'Accept': 'application/json'}
        )
        self.assertEqual(200, response.status_code)
        self.assertEqual(3, len(response.json))
        response = self.testapp.get(
            '/conceptschemes/TREES/c'
            '?match=http://id.python.org/different/types/of/trees/nr/1/the/larch',
            headers={'Accept': 'application/json'}
        )
        self.assertEqual(200, response.status_code)
        self.assertEqual(
            [
                {
                    'id': '1',
                    'uri': 'urn:x-skosprovider:trees/1',
                    'type': 'concept',
                    'label': 'De Lariks',
                    '@context': 'http://localhost/jsonld/context/skos'
                }
            ],
            response.json
        )


class CacheFunctionalTests(FunctionalTests):
    def _get_default_headers(self):
        return {'Accept': 'application/json'}

    def test_create_cache(self):
        # clear entire cache before start
        invalidate_cache_response = self.testapp.get('/admin/tree/invalidate')
        self.assertEqual('200 OK', invalidate_cache_response.status)

        tree_response = self.testapp.get('/conceptschemes/MATERIALS/tree?_LOCALE_=nl',
                                         headers=self._get_default_headers())
        self.assertEqual('200 OK', tree_response.status)
        self.assertIsNotNone(tree_response.json)

        cached_tree_response = self.testapp.get(
            '/conceptschemes/MATERIALS/tree?_LOCALE_=nl',
            headers=self._get_default_headers())
        self.assertEqual('200 OK', cached_tree_response.status)
        self.assertIsNotNone(cached_tree_response.json)

        self.assertEqual(tree_response.json, cached_tree_response.json)

    def test_auto_invalidate_cache(self):
        tree_region.configure('dogpile.cache.memory',
                              expiration_time=7000,
                              arguments={'cache_size': 5000},
                              replace_existing_backend=True)
        list_region.configure('dogpile.cache.memory',
                              expiration_time=7000,
                              arguments={'cache_size': 5000},
                              replace_existing_backend=True)
        # clear entire cache before start
        invalidate_cache_response = self.testapp.get('/admin/tree/invalidate')
        self.assertEqual('200 OK', invalidate_cache_response.status)

        tree_response = self.testapp.get('/conceptschemes/MATERIALS/tree?_LOCALE_=nl',
                                         headers=self._get_default_headers())
        cached_tree_response = self.testapp.get(
            '/conceptschemes/MATERIALS/tree?_LOCALE_=nl',
            headers=self._get_default_headers())
        self.assertEqual(tree_response.json, cached_tree_response.json)

        delete_response = self.testapp.delete('/conceptschemes/MATERIALS/c/31',
                                              headers=self._get_default_headers())
        self.assertEqual('200 OK', delete_response.status)
        self.assertIsNotNone(delete_response.json['id'])

        tree_response2 = self.testapp.get('/conceptschemes/MATERIALS/tree?_LOCALE_=nl',
                                          headers=self._get_default_headers())
        self.assertNotEqual(tree_response.json, tree_response2.json)

        cached_tree_response2 = self.testapp.get(
            '/conceptschemes/MATERIALS/tree?_LOCALE_=nl',
            headers=self._get_default_headers())
        self.assertEqual(tree_response2.json, cached_tree_response2.json)

        tree_region.configure('dogpile.cache.null', replace_existing_backend=True)
        list_region.configure('dogpile.cache.null', replace_existing_backend=True)


class RdfFunctionalTests(FunctionalTests):

    def test_void(self):
        rdf_response = self.testapp.get('/void.ttl')
        self.assertEqual('200 OK', rdf_response.status)
        self.assertEqual('text/turtle', rdf_response.content_type)

    def test_rdf_full_xml(self):
        rdf_response = self.testapp.get('/conceptschemes/MATERIALS/c',
                                        headers={'Accept': 'application/rdf+xml'})
        self.assertEqual('200 OK', rdf_response.status)
        self.assertEqual('application/rdf+xml', rdf_response.content_type)

    def test_rdf_full_xml_ext(self):
        rdf_response = self.testapp.get('/conceptschemes/MATERIALS/c.rdf')
        self.assertEqual('200 OK', rdf_response.status)
        self.assertEqual('application/rdf+xml', rdf_response.content_type)

    def test_rdf_full_turtle(self):
        ttl_response = self.testapp.get('/conceptschemes/MATERIALS/c',
                                        headers={'Accept': 'text/turtle'})
        self.assertEqual('200 OK', ttl_response.status)
        self.assertEqual('text/turtle', ttl_response.content_type)

    def test_rdf_full_turtle_ext(self):
        ttl_response = self.testapp.get('/conceptschemes/MATERIALS/c.ttl')
        self.assertEqual('200 OK', ttl_response.status)
        self.assertEqual('text/turtle', ttl_response.content_type)

    def test_rdf_conceptscheme_xml(self):
        rdf_response = self.testapp.get('/conceptschemes/MATERIALS',
                                        headers={'Accept': 'application/rdf+xml'})
        self.assertEqual('200 OK', rdf_response.status)
        self.assertEqual('application/rdf+xml', rdf_response.content_type)

    def test_rdf_conceptscheme_xml_ext(self):
        rdf_response = self.testapp.get('/conceptschemes/MATERIALS.rdf')
        self.assertEqual('200 OK', rdf_response.status)
        self.assertEqual('application/rdf+xml', rdf_response.content_type)

    def test_rdf_conceptscheme_turtle(self):
        ttl_response = self.testapp.get('/conceptschemes/MATERIALS',
                                        headers={'Accept': 'text/turtle'})
        self.assertEqual('200 OK', ttl_response.status)
        self.assertEqual('text/turtle', ttl_response.content_type)

    def test_rdf_conceptscheme_turtle_ext(self):
        ttl_response = self.testapp.get('/conceptschemes/MATERIALS.ttl')
        self.assertEqual('200 OK', ttl_response.status)
        self.assertEqual('text/turtle', ttl_response.content_type)

    def test_rdf_conceptscheme_jsonld(self):
        res = self.testapp.get('/conceptschemes/MATERIALS',
                               headers={'Accept': 'application/ld+json'})
        self.assertEqual('200 OK', res.status)
        self.assertEqual('application/ld+json', res.content_type)

    def test_rdf_conceptscheme_jsonld_ext(self):
        res = self.testapp.get('/conceptschemes/MATERIALS.jsonld')
        self.assertEqual('200 OK', res.status)
        self.assertEqual('application/ld+json', res.content_type)

    def test_rdf_individual_jsonld(self):
        res = self.testapp.get('/conceptschemes/MATERIALS/c/1',
                               headers={'Accept': 'application/ld+json'})
        self.assertEqual('200 OK', res.status)
        self.assertEqual('application/ld+json', res.content_type)

    def test_rdf_individual_jsonld_ext(self):
        res = self.testapp.get('/conceptschemes/MATERIALS/c/1.jsonld')
        self.assertEqual('200 OK', res.status)
        self.assertEqual('application/ld+json', res.content_type)

    def test_rdf_individual_xml(self):
        rdf_response = self.testapp.get('/conceptschemes/MATERIALS/c/1',
                                        headers={'Accept': 'application/rdf+xml'})
        self.assertEqual('200 OK', rdf_response.status)
        self.assertEqual('application/rdf+xml', rdf_response.content_type)

    def test_rdf_individual_xml_ext(self):
        rdf_response = self.testapp.get('/conceptschemes/MATERIALS/c/1.rdf')
        self.assertEqual('200 OK', rdf_response.status)
        self.assertEqual('application/rdf+xml', rdf_response.content_type)

    def test_rdf_individual_turtle(self):
        ttl_response = self.testapp.get('/conceptschemes/MATERIALS/c/1',
                                        headers={'Accept': 'text/turtle'})
        self.assertEqual('200 OK', ttl_response.status)
        self.assertEqual('text/turtle', ttl_response.content_type)

    def test_rdf_individual_turtle_ext(self):
        ttl_response = self.testapp.get('/conceptschemes/MATERIALS/c/1.ttl')
        self.assertEqual('200 OK', ttl_response.status)
        self.assertEqual('text/turtle', ttl_response.content_type)

    def test_rdf_individual_jsonld_ext_manual(self):
        res = self.testapp.get('/conceptschemes/manual-ids/c/manual-1.jsonld')
        self.assertEqual('200 OK', res.status)
        self.assertEqual('application/ld+json', res.content_type)

    def test_rdf_individual_xml_ext_manual(self):
        rdf_response = self.testapp.get('/conceptschemes/manual-ids/c/manual-2.rdf')
        self.assertEqual('200 OK', rdf_response.status)
        self.assertEqual('application/rdf+xml', rdf_response.content_type)

    def test_rdf_individual_turtle_manual(self):
        ttl_response = self.testapp.get('/conceptschemes/manual-ids/c/manual-1.ttl')
        self.assertEqual('200 OK', ttl_response.status)
        self.assertEqual('text/turtle', ttl_response.content_type)

    def test_rdf_individual_turtle_manual_uri(self):
        ttl_response = self.testapp.get(
            '/conceptschemes/manual-ids/c/http://id.manual.org/manual/68.ttl')
        self.assertEqual('200 OK', ttl_response.status)
        self.assertEqual('text/turtle', ttl_response.content_type)

    def test_rdf_individual_not_found(self):
        res = self.testapp.get('/conceptschemes/TREES/c/test.ttl',
                               headers={'Accept': 'text/turtle'}, status=404,
                               expect_errors=True)
        self.assertEqual('404 Not Found', res.status)


class ListFunctionalTests(FunctionalTests):
    def test_labeltypes_list(self):
        labeltypeslist_res = self.testapp.get('/labeltypes')
        self.assertEqual('200 OK', labeltypeslist_res.status)
        self.assertEqual('application/json', labeltypeslist_res.content_type)
        self.assertIsNotNone(labeltypeslist_res.json)
        self.assertEqual(4, len(labeltypeslist_res.json))

    def test_notetypes_list(self):
        labeltypeslist_res = self.testapp.get('/notetypes')
        self.assertEqual('200 OK', labeltypeslist_res.status)
        self.assertEqual('application/json', labeltypeslist_res.content_type)
        self.assertIsNotNone(labeltypeslist_res.json)



"""unique concept.id + conceptscheme.id

Revision ID: 79ff53a30228
Revises: 3e9675b35dfc
Create Date: 2020-09-14 12:23:22.546226

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "79ff53a30228"
down_revision = "3e9675b35dfc"


def upgrade():
    with op.batch_alter_table("concept") as batch_op:
        batch_op.create_unique_constraint(
            "uq_concept_concept_id_conceptscheme_id", ["concept_id", "conceptscheme_id"]
        )


def downgrade():
    with op.batch_alter_table("concept") as batch_op:
        batch_op.drop_constraint("uq_concept_concept_id_conceptscheme_id")



import unittest

from pyramid import testing
from pyramid.httpexceptions import HTTPMethodNotAllowed
from skosprovider.providers import DictionaryProvider
from skosprovider.skos import Collection as SkosCollection
from skosprovider.skos import Concept as SkosConcept
from skosprovider_sqlalchemy.models import Collection
from skosprovider_sqlalchemy.models import Concept
from skosprovider_sqlalchemy.models import ConceptScheme
from skosprovider_sqlalchemy.models import Label
from skosprovider_sqlalchemy.models import Match
from skosprovider_sqlalchemy.models import MatchType
from skosprovider_sqlalchemy.models import Note
from skosprovider_sqlalchemy.models import Source
from skosprovider_sqlalchemy.providers import SQLAlchemyProvider

from atramhasis.utils import from_thing
from atramhasis.utils import internal_providers_only
from atramhasis.utils import update_last_visited_concepts

species = {
    'id': 3,
    'uri': 'http://id.trees.org/3',
    'labels': [
        {'type': 'prefLabel', 'language': 'en', 'label': 'Trees by species'},
        {'type': 'prefLabel', 'language': 'nl', 'label': 'Bomen per soort'}
    ],
    'type': 'collection',
    'members': ['1', '2'],
    'notes': [
        {
            'type': 'editorialNote',
            'language': 'en',
            'note': 'As seen in How to Recognise Different Types of Trees from Quite a Long Way Away.'
        }
    ]
}


class TestFromThing(unittest.TestCase):
    def setUp(self):
        conceptscheme = ConceptScheme()
        conceptscheme.uri = 'urn:x-atramhasis-demo'
        conceptscheme.id = 1
        self.concept = Concept()
        self.concept.id = 11
        self.concept.concept_id = 101
        self.concept.uri = 'urn:x-atramhasis-demo:TREES:101'
        self.concept.conceptscheme_id = 1
        self.concept.conceptscheme = conceptscheme

        notes = []
        note = Note(note='test note', notetype_id='example', language_id='en')
        note2 = Note(note='note def', notetype_id='definition', language_id='en')
        notes.append(note)
        notes.append(note2)
        self.concept.notes = notes

        labels = []
        label = Label(label='een label', labeltype_id='prefLabel', language_id='nl')
        label2 = Label(label='other label', labeltype_id='altLabel', language_id='en')
        label3 = Label(label='and some other label', labeltype_id='altLabel', language_id='en')
        labels.append(label)
        labels.append(label2)
        labels.append(label3)
        self.concept.labels = labels

        sources = []
        source = Source(citation='Kinsella S. & Carlisle P. 2015: Alice.')
        sources.append(source)
        self.concept.sources = sources

        matches = []
        match1 = Match()
        match1.uri = 'urn:test'
        match1.concept = self.concept
        match1.matchtype = MatchType(name='closeMatch', description='')
        match2 = Match()
        match2.uri = 'urn:test'
        match2.concept = self.concept
        match2.matchtype = MatchType(name='closeMatch', description='')
        matches.append(match1)
        matches.append(match2)
        self.matches = matches

        self.collection = Collection()
        self.collection.id = 12
        self.collection.concept_id = 102
        self.collection.uri = 'urn:x-atramhasis-demo:TREES:102'
        self.collection.conceptscheme_id = 1
        self.collection.conceptscheme = conceptscheme

    def test_thing_to_concept(self):
        skosconcept = from_thing(self.concept)
        self.assertTrue(isinstance(skosconcept, SkosConcept))
        self.assertEqual(skosconcept.id, 101)
        self.assertEqual(len(skosconcept.labels), 3)
        self.assertEqual(len(skosconcept.notes), 2)
        self.assertEqual(len(skosconcept.sources), 1)
        self.assertEqual(skosconcept.uri, 'urn:x-atramhasis-demo:TREES:101')

    def test_thing_to_collection(self):
        skoscollection = from_thing(self.collection)
        self.assertTrue(isinstance(skoscollection, SkosCollection))
        self.assertEqual(skoscollection.id, 102)
        self.assertEqual(skoscollection.uri, 'urn:x-atramhasis-demo:TREES:102')


class DummyViewClassForTest:

    def __init__(self):
        self.provider = None
        self.dummy = None

    def all_providers(self, dummy):
        self.dummy = dummy

    @internal_providers_only
    def internal_providers(self, dummy):
        self.dummy = dummy


class TestInternalProviderOnly(unittest.TestCase):

    def setUp(self):
        self.dummy = DummyViewClassForTest()

    def test_all_providers(self):
        self.dummy.provider = DictionaryProvider(list=[species], metadata={'id': 'Test'})
        self.dummy.all_providers('ok')
        self.assertEqual(self.dummy.dummy, 'ok')

    def test_internal_providers(self):
        self.dummy.provider = SQLAlchemyProvider(metadata={'id': 'Test', 'conceptscheme_id': 1},
                                                 session=None)
        self.dummy.internal_providers('ok')
        self.assertEqual(self.dummy.dummy, 'ok')

    def test_external_providers(self):
        self.dummy.provider = SQLAlchemyProvider(metadata={'id': 'Test', 'conceptscheme_id': 1,
                                                           'subject': ['external']},
                                                 session=None)
        self.assertRaises(HTTPMethodNotAllowed, self.dummy.internal_providers, 'ok')
        self.assertIsNone(self.dummy.dummy)

    def test_no_sqlalchemyprovider(self):
        self.dummy.provider = DictionaryProvider(list=[species], metadata={'id': 'Test'})
        self.assertRaises(HTTPMethodNotAllowed, self.dummy.internal_providers, 'ok')
        self.assertIsNone(self.dummy.dummy)


class TestUpdateLastVisitedConceptsProviderOnly(unittest.TestCase):

    def setUp(self):
        self.config = testing.setUp()
        self.request = testing.DummyRequest()
        self.request.session = {}

    def tearDown(self):
        testing.tearDown()

    def test_update_last_visited_concepts(self):
        c = Concept()
        c.id = 2
        c.labels = [Label('test', language_id='en-us')]
        update_last_visited_concepts(self.request, {'label': c.label(), 'url': f'http://test.test/{55}'})
        c = Concept()
        c.id = 33
        c.labels = [Label('test', language_id='nl-be')]
        update_last_visited_concepts(self.request, {'label': c.label(), 'url': f'http://test.test/{2}'})
        self.assertEqual(2, len(self.request.session['last_visited']))

    def test_update_last_visited_concepts_max(self):
        for concept_id in range(50):
            c = Concept()
            c.id = concept_id
            c.labels = [Label('test', language_id='en-us')]
            update_last_visited_concepts(
                self.request, {'label': c.label(), 'url': f'http://test.test/{concept_id}'}
            )
        self.assertEqual(4, len(self.request.session['last_visited']))
        last = self.request.session['last_visited'].pop()
        self.assertEqual('http://test.test/49', last['url'])

    def test_no_double_last_visited_concepts(self):
        c = Concept()
        c.id = 2
        c.labels = [Label('test', language_id='en-us')]
        update_last_visited_concepts(self.request, {'label': c.label(), 'url': f'http://test.test/{55}'})
        update_last_visited_concepts(self.request, {'label': c.label(), 'url': f'http://test.test/{55}'})
        c = Concept()
        c.id = 33
        c.labels = [Label('test', language_id='nl-be')]
        update_last_visited_concepts(self.request, {'label': c.label(), 'url': f'http://test.test/{2}'})
        self.assertEqual(2, len(self.request.session['last_visited']))



"""
Module containing mapping functions used by Atramhasis.
"""

from skosprovider_sqlalchemy.models import Collection
from skosprovider_sqlalchemy.models import Concept
from skosprovider_sqlalchemy.models import ConceptScheme
from skosprovider_sqlalchemy.models import Label
from skosprovider_sqlalchemy.models import Match
from skosprovider_sqlalchemy.models import Note
from skosprovider_sqlalchemy.models import Source
from skosprovider_sqlalchemy.providers import SQLAlchemyProvider
from sqlalchemy.exc import NoResultFound

from atramhasis.data.models import ExpandStrategy
from atramhasis.data.models import IDGenerationStrategy
from atramhasis.data.models import Provider


def is_html(value):
    """
    Check if a value has html inside. Only tags checked <strong> <em> <a>.

    :param value: a string
    :return: a boolean (True, HTML present | False, no HTML present)
    """
    tag_list = ['<strong>', '<em>', '<a>', '</strong>', '</em>', '</a>', '<a']
    return any(tag in value for tag in tag_list)


def map_concept(concept, concept_json, skos_manager):
    """
    Map a concept from json to the database.

    :param skosprovider_sqlalchemy.models.Thing concept: A concept or
        collection as known to the database.
    :param dict concept_json: A dict representing the json sent to our REST
        service.
    :param skos_manager: A skos_manager to acces db operations
    :returns: The :class:`skosprovider_sqlalchemy.models.Thing` enhanced
        with the information from the json object.
    """
    concept_json_type = concept_json.get('type', None)
    if concept.type != concept_json_type:

        if concept_json_type == 'concept':
            members = concept.members
            concept = skos_manager.change_type(
                concept,
                concept.concept_id,
                concept.conceptscheme_id,
                concept_json_type,
                concept.uri
            )
            for member in members:
                if member.type == 'concept':
                    concept.narrower_concepts.add(member)
                elif member.type == 'collection':
                    concept.narrower_collections.add(member)
        elif concept_json_type == 'collection':
            narrower_concepts = concept.narrower_concepts
            narrower_collections = concept.narrower_collections
            concept = skos_manager.change_type(
                concept,
                concept.concept_id,
                concept.conceptscheme_id,
                concept_json_type,
                concept.uri
            )
            for narrower_concept in narrower_concepts:
                concept.members.add(narrower_concept)
            for narrower_collection in narrower_collections:
                concept.members.add(narrower_collection)
    elif concept_json_type == 'collection':
        concept.members.clear()
    elif concept_json_type == 'concept':
        concept.narrower_collections.clear()
        concept.narrower_concepts.clear()
    if concept.type in ('concept', 'collection'):
        concept.labels[:] = []
        labels = concept_json.get('labels', [])
        for label in labels:
            label = Label(
                label=label.get('label', ''),
                labeltype_id=label.get('type', ''),
                language_id=label.get('language', ''),
            )
            concept.labels.append(label)
        concept.notes[:] = []
        notes = concept_json.get('notes', [])
        for n in notes:
            note = Note(note=n.get('note', ''), notetype_id=n.get('type', ''), language_id=n.get('language', ''))
            if is_html(note.note):
                note.markup = 'HTML'
            concept.notes.append(note)
        concept.sources[:] = []
        sources = concept_json.get('sources', [])
        for s in sources:
            source = Source(citation=s.get('citation', ''))
            if is_html(source.citation):
                source.markup = 'HTML'
            concept.sources.append(source)

        concept.member_of.clear()
        member_of = concept_json.get('member_of', [])
        for memberof in member_of:
            try:
                memberof_collection = skos_manager.get_thing(
                    concept_id=memberof['id'],
                    conceptscheme_id=concept.conceptscheme_id)
            except NoResultFound:
                memberof_collection = Collection(concept_id=memberof['id'], conceptscheme_id=concept.conceptscheme_id)
            concept.member_of.add(memberof_collection)

        if concept.type == 'concept':
            concept.related_concepts.clear()
            related = concept_json.get('related', [])
            for related in related:
                try:
                    related_concept = skos_manager.get_thing(
                        concept_id=related['id'],
                        conceptscheme_id=concept.conceptscheme_id)
                except NoResultFound:
                    related_concept = Concept(concept_id=related['id'], conceptscheme_id=concept.conceptscheme_id)
                concept.related_concepts.add(related_concept)

            concept.broader_concepts.clear()
            broader = concept_json.get('broader', [])
            for broader in broader:
                try:
                    broader_concept = skos_manager.get_thing(
                        concept_id=broader['id'],
                        conceptscheme_id=concept.conceptscheme_id)
                except NoResultFound:
                    broader_concept = Concept(concept_id=broader['id'], conceptscheme_id=concept.conceptscheme_id)
                concept.broader_concepts.add(broader_concept)
            narrower = concept_json.get('narrower', [])
            for narrower in narrower:
                try:
                    narrower_concept = skos_manager.get_thing(
                        concept_id=narrower['id'],
                        conceptscheme_id=concept.conceptscheme_id)
                except NoResultFound:
                    narrower_concept = Concept(concept_id=narrower['id'], conceptscheme_id=concept.conceptscheme_id)
                concept.narrower_concepts.add(narrower_concept)

            matches = []
            matchdict = concept_json.get('matches', {})
            for match_type, uris in matchdict.items():
                db_type = match_type + "Match"
                match_type = skos_manager.get_match_type(db_type)
                for uri in uris:
                    concept_id = concept_json.get('id', -1)
                    try:
                        match = skos_manager.get_match(uri=uri, matchtype_id=match_type.name,
                                                       concept_id=concept_id)
                    except NoResultFound:
                        match = Match()
                        match.matchtype = match_type
                        match.uri = uri
                    matches.append(match)
            concept.matches = matches

            narrower_collections = concept_json.get('subordinate_arrays', [])
            for narrower in narrower_collections:
                try:
                    narrower_collection = skos_manager.get_thing(
                        concept_id=narrower['id'],
                        conceptscheme_id=concept.conceptscheme_id)
                except NoResultFound:
                    narrower_collection = Collection(concept_id=narrower['id'],
                                                     conceptscheme_id=concept.conceptscheme_id)
                concept.narrower_collections.add(narrower_collection)

        if concept.type == 'collection':
            members = concept_json.get('members', [])
            for member in members:
                try:
                    member_concept = skos_manager.get_thing(
                        concept_id=member['id'],
                        conceptscheme_id=concept.conceptscheme_id)
                except NoResultFound:
                    member_concept = Concept(concept_id=member['id'], conceptscheme_id=concept.conceptscheme_id)
                concept.members.add(member_concept)

            concept.broader_concepts.clear()
            broader_concepts = concept_json.get('superordinates', [])
            for broader in broader_concepts:
                try:
                    broader_concept = skos_manager.get_thing(
                        concept_id=broader['id'],
                        conceptscheme_id=concept.conceptscheme_id)
                except NoResultFound:
                    broader_concept = Concept(concept_id=broader['id'], conceptscheme_id=concept.conceptscheme_id)
                concept.broader_concepts.add(broader_concept)
            if 'infer_concept_relations' in concept_json:
                concept.infer_concept_relations = concept_json['infer_concept_relations']
    return concept


def map_conceptscheme(conceptscheme, conceptscheme_json):
    """
    Map a conceptscheme from json to the database.

    :param skosprovider_sqlalchemy.models.ConceptScheme conceptscheme: A conceptscheme as known to the database.
    :param dict conceptscheme_json: A dict representing the json sent to our REST
        service.
    :returns: The :class:`skosprovider_sqlalchemy.models.ConceptScheme` enhanced
        with the information from the json object.
    """
    conceptscheme.labels[:] = []
    labels = conceptscheme_json.get('labels', [])
    for label in labels:
        label = Label(
            label=label.get('label', ''),
            labeltype_id=label.get('type', ''),
            language_id=label.get('language', ''),
        )
        conceptscheme.labels.append(label)
    conceptscheme.notes[:] = []
    notes = conceptscheme_json.get('notes', [])
    for n in notes:
        note = Note(note=n.get('note', ''), notetype_id=n.get('type', ''), language_id=n.get('language', ''))
        conceptscheme.notes.append(note)
    conceptscheme.sources[:] = []
    sources = conceptscheme_json.get('sources', [])
    for s in sources:
        source = Source(citation=s.get('citation', ''))
        conceptscheme.sources.append(source)
    return conceptscheme


def map_provider(provider_json: dict, provider: Provider = None) -> Provider:
    """
    Create a atramhasis.data.models.Provider from json data.

    An existing provider can optionally be passed. When passed this one will be updated.

    :param provider_json: JSON data as a dict.
    :param provider: A provider which will be updated with the JSON data. When None
       a new Provider instance will be returned.
    :return: A Provider set with data from the JSON.
    """
    if provider is None:
        # Only executed on creation.
        provider = Provider()
        provider.conceptscheme = ConceptScheme(uri=provider_json['conceptscheme_uri'])
        provider.id = provider_json.get("id")

    provider.meta = provider_json.get("metadata") or {}
    provider.default_language = provider_json.get("default_language")
    provider.force_display_language = provider_json.get("force_display_language")
    provider.id_generation_strategy = IDGenerationStrategy[
        provider_json.get("id_generation_strategy") or 'NUMERIC'
    ]
    provider.subject = provider_json.get("subject") or []
    provider.uri_pattern = provider_json["uri_pattern"]
    provider.expand_strategy = ExpandStrategy[
        (provider_json.get("expand_strategy") or 'RECURSE').upper()
    ]
    return provider
